
/* $Id: cpdcons.c,v 3.15 1999/11/08 02:01:34 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <string.h>

/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"
#include "cpd_proto.h"

void	read_name();


/* a main routine to test the reading and writing */

main()
{
	char app[80];
	char cline[80];
	char command[80];

	char *option;

	int bufid;
	int cc;
	int cpdtid;
	int qflag = 0;
	int restart;

	/* Initialize CUMULVS */

	stv_myid( &STV_MYID );

	STV_USERTAG = 10;

	if ( stv_verify_system_contexts( stvTrue, stvTrue ) < 0 )
		exit( -1 );

	/* Set Up Message Handlers */

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			stv_protoConfigCpd, stvConfigCpd,
			stvAny, STV_USERTAG, STV_APP_CTX );

	stv_add_handler( &STVHANDLERS, &nSTVHANDLERS,
			cpd_protoCpdFileList, stvCpdFileList,
			stvAny, stvAny, STV_CPD_CTX );

	/* Contact / Start Local CPD */
	
	if ( (cpdtid = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "main", stvStatusSysFail,
				"Cannot Find or (Re)Start Cpd...  Bailing." );
		exit( -1 );
	}

	/* Command Loop */

 	while( !qflag )
	{
		printf( "cpd> " );
		read_name( cline );	
		sscanf( cline, "%s", command );

		if ( !strcmp( command, "quit" ) )
			qflag = 1;	

		if ( !strcmp( command, "restart" ) )
		{
			sscanf( cline, "%s %s\n", command, app );

			stv_init_msg_buf( stvDataDefault, &bufid );

			cc = stvMsgOp( stvRestart );
			stv_pack_int( bufid, &cc, 1, 1 );

			/* dummy fgid, for message header consistency */
			cc = -1;
			stv_pack_int( bufid, &cc, 1, 1 );

			stv_pack_str( bufid, app );

			cc = 1;
			stv_pack_int( bufid, &cc, 1, 1 );

			stv_send_msg( bufid, cpdtid, stvCpdTag, STV_CPD_CTX );

			stv_free_msg_buf( bufid );
		}

		if ( !strcmp( command, "list" ) )
			cpd_list_local_files();

		if ( !strcmp( command, "boot" ) )
			cpd_bootstrap();

		if ( !strcmp( command, "help" ) )
		{
			printf( "Commands: \n" );
			printf( "\thelp - this message.\n" );
			printf( "\tquit - leave the cpd console.\n" );
			printf( "\tlist - list local checkpoint files.\n" );
			printf( "\tboot - bootstrap an application from a " );
			printf( "checkpoint.\n" );
			printf( "\trestart <app> - restart the running " );
			printf( "application.\n" );
		}
	}			
}		


void
read_name( str )
char *str;
{
    while ( (*str = getchar()) != '\n' )
        str++;

    *str = '\0';
}


/*list local chkpt files (actually get my cpd to do it) */
int
cpd_list_local_files()
{
	STV_BUFPTR msg;
	int bufid;
	int cc, cpdtid;
	
	int dummy;

	if ( (cpdtid = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "cpd_list_local_files", stvStatusSysFail,
				"Cannot Find or (Re)Start Cpd...  Bailing." );
		exit( -1 );
	}

	stv_init_msg_buf( stvDataDefault, &bufid );

	cc = stvMsgOp( stvCpdListFiles );
	stv_pack_int( bufid, &cc, 1, 1 );

	/* dummy fgid, for message header consistency */
	cc = -1;
	stv_pack_int( bufid, &cc, 1, 1 );

	stv_send_msg( bufid, cpdtid, stvCpdTag, STV_CPD_CTX );

	stv_free_msg_buf( bufid );

	printf( "Contacting Local Checkpointing Daemon (Cpd).\n");

	if ( stv_handleMessages( stvCpdFileList,
			cpdtid, stvAny, STV_CPD_CTX, stvMsgBlock ) <= 0 )
    {
		STV_ERROR( "cpd_list_local_files", stvStatusSysFail,
				"Cpd Exited While I Was Waiting For File Listing." );
		exit( -1 );
    }

	return( stvStatusOk );
}


/* hack so that I can use cpd_mp.o,this really sucks! XXX */
cpd_is_peer( tid )	
int tid;
{
	return( stvFalse );
}


/* Bootstrap an application */
int
cpd_bootstrap()
{
	STV_HOSTPTR hostp;

	char cline[256];
	char name[stvMaxNameLen];

	int bufid;
	int cc;
	int cpdtid;
	int i;
	int nhosts;
	int nspawn;

	if ( (cpdtid = stv_find_cpd()) < 0 )
	{
		STV_ERROR( "cpd_bootstrap", stvStatusSysFail,
				"Cannot Find or (Re)Start Cpd...  Bailing." );
		exit( -1 );
	}

	if ( !stv_host_config( &hostp, &nhosts ) )
	{
		STV_ERROR( "cpd_bootstrap", stvStatusSysFail,
				"Cannot Get Host Config...  Bailing." );
		exit( -1 );
	}

	cpd_list_local_files();

	printf( "Checkpoint file to bootstrap: " );
	fflush( stdout );
	read_name( cline );
	sscanf( cline, "%s", name );
	
	stv_init_msg_buf( stvDataDefault, &bufid );

	cc = stvMsgOp( stvCpdBootstrap );
	stv_pack_int( bufid, &cc, 1, 1 );

	/* dummy fgid, for message header consistency */
	cc = -1;
	stv_pack_int( bufid, &cc, 1, 1 );

	stv_pack_str( bufid, name );

	stv_pack_int( bufid, &nhosts, 1, 1 );

	printf( "How many total tasks ? ");
	fflush( stdout );
	read_name(cline);
	sscanf(cline, "%d", &nspawn);

	stv_pack_int( bufid, &nspawn, 1, 1 );

	for ( i=0 ; i < nhosts ; i++ )
	{
		printf( "How many on %s ? ", hostp[i].name );
		fflush( stdout );
		read_name(cline);
		sscanf(cline, "%d", &nspawn);

		stv_pack_int( bufid, &(hostp[i].tid), 1, 1 );

		stv_pack_int( bufid, &nspawn, 1, 1 );
	}

	FREE( hostp );

	stv_send_msg( bufid, cpdtid, stvCpdTag, STV_CPD_CTX );

	stv_free_msg_buf( bufid );

	return( stvStatusOk );
}

