#
# $Id: procs.tcl,v 3.0 1996/08/23 22:06:30 kohl Exp $
#


#
#                       CUMULVS version 1.0:
#       Computational Steering and Interactive Visualization
#                    in Distributed Applications
#            Oak Ridge National Laboratory, Oak Ridge TN.
#       Authors:  James Arthur Kohl and Philip M. Papadopoulos
#                    (C) 1996 All Rights Reserved
#
#                               NOTICE
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted
#  provided that the above copyright notice appear in all copies and
#  that both the copyright notice and this permission notice appear
#  in supporting documentation.
#
#  Neither the Institution, Oak Ridge National Laboratory, nor the
#  Authors make any representations about the suitability of this
#  software for any purpose.  This software is provided ``as is''
#  without express or implied warranty.
#
#  CUMULVS was funded by the U.S. Department of Energy.
#


puts -nonewline "\[procs.tcl\]"
flush stdout


#
# FlyMe TCL Procs
#


proc handleMode { } \
{
	#debug_proc handleMode entry

	global play_mode

	layout_main_panel

	set_play_mode $play_mode

	#debug_proc handleMode exit
}


proc handleOpts { } \
{
	#debug_proc handleOpts entry

	global auto_steer
	global auto_zoom

	set_opts $auto_steer $auto_zoom

	#debug_proc handleOpts exit
}


proc zoom_factor_handle { } \
{
	#debug_proc zoom_factor_handle entry

	set FM .flyme

	set zf [ $FM.zoom_factor get ]

	set_zoom_factor $zf

	#debug_proc zoom_factor_handle exit
}


proc radar_range_handle { } \
{
	#debug_proc radar_range_handle entry

	set FM .flyme

	set zf [ $FM.radar_range get ]

	if { $zf != "" } \
		{ set_radar_range $zf }

	#debug_proc radar_range_handle exit
}


proc resize_main_panel { } \
{
	#debug_proc resize_main_panel entry

	global main_height
	global main_width

	set ht [ winfo height .flyme ]
	set wt [ winfo width .flyme ]

	if { $main_width != $wt || $main_height != $ht } \
	{
		set main_height $ht
		set main_width $wt

		layout_main_panel

		resize_stv
	}

	#debug_proc resize_main_panel exit
}


proc layout_main_panel { } \
{
	#debug_proc layout_main_panel entry

	global border_space
	global frame_border
	global flyme_height
	global flyme_width
	global flyme_ssize
	global flyme_hsize
	global main_height
	global main_width
	global flyme_size
	global row_height
	global play_mode

	global FM
	global FM_C
	global FM_R
	global FM_I
	global FM_IF
	global FM_IS
	global FM_IH
	global FM_IH_N
	global FM_IH_S
	global FM_IH_E
	global FM_IH_W

	#
	# Place Message
	#

	set msg_y [ expr [ below .flyme.title ] + (3 * $border_space / 2) ]

	set msg_wt [ expr $main_width - (2 * $border_space) ]

	place $FM.message -x $border_space -y $msg_y -width $msg_wt

	update

	#
	# Layout Main Panel
	#

	set y [ expr [ below .flyme.message ] + $border_space ]

	place $FM.unzoom -x $border_space -y $y

	update

	place $FM.full_unzoom -x [ right $FM.unzoom ] -y $y

	update

	place $FM.quit -x [ right $FM.full_unzoom ] -y $y

	update

	set x [ expr $border_space + 48 ]

	set y [ expr [ below $FM.quit ] + $border_space + 48 ]

	place $FM.up -x $x -y [ expr $y - 32 ] -anchor center

	place $FM.down -x $x -y [ expr $y + 32 ] -anchor center

	place $FM.left -x [ expr $x - 32 ] -y $y -anchor center

	place $FM.right -x [ expr $x + 32 ] -y $y -anchor center

	place $FM.ne -x [ expr $x + 32 ] -y [ expr $y - 32 ] -anchor center

	place $FM.nw -x [ expr $x - 32 ] -y [ expr $y - 32 ] -anchor center

	place $FM.sw -x [ expr $x - 32 ] -y [ expr $y + 32 ] -anchor center

	place $FM.se -x [ expr $x + 32 ] -y [ expr $y + 32 ] -anchor center

	set x [ expr $x + 48 + $border_space ]

	set my [ expr [ below $FM.quit ] + $border_space ]

	place $FM.monitor -x $x -y $my -anchor nw

	update

	place $FM.play -x $x -y [ below $FM.monitor ] -anchor nw

	if { $play_mode == "Play" } \
	{
		set x [ expr [ right $FM.monitor ] + $border_space ]

		place $FM.auto_zoom -x $x -y $my -anchor nw

		update

		place $FM.auto_steer -x $x -y [ below $FM.auto_zoom ] -anchor nw

		update

		set x [ expr [ right $FM.auto_zoom ] + $border_space ]

		place $FM.zoom_factor_label -x $x -y $my -anchor nw

		update

		set ex [ right $FM.zoom_factor_label ]

		set wt [ expr $main_width - $ex - $border_space ]

		place $FM.zoom_factor -x $ex -y $my -anchor nw -width $wt

		set ry [ expr [ below $FM.zoom_factor_label ] + $border_space ]

		place $FM.radar_range_label -x $x -y $ry -anchor nw

		update

		set ex [ right $FM.radar_range_label ]

		set wt [ expr $main_width - $ex - $border_space ]

		place $FM.radar_range -x $ex -y $ry -anchor nw -width $wt

		set sy [ expr [ below $FM.radar_range_label ] + $border_space ]

		place $FM.score_label -x $x -y $sy -anchor nw
	} \
\
	else \
	{
		place forget $FM.auto_steer
		place forget $FM.auto_zoom

		place forget $FM.zoom_factor_label
		place forget $FM.zoom_factor

		place forget $FM.radar_range_label
		place forget $FM.radar_range

		place forget $FM.score_label
	}

	set y [ expr $y + 64 + $border_space ]

	set height_left [ expr $main_height - $y - $border_space ]
	set width_left [ expr $main_width - (2 * $border_space) ]

	if { $play_mode == "Play" } \
	{
		set eff_width [ expr ( 2 * $width_left ) / 3 ]

		if { $height_left < $eff_width } \
		{
			set flyme_height $height_left
			set flyme_width $height_left

			set flyme_size [ expr $height_left / 2 ]
		} \
\
		else \
		{
			set flyme_height $eff_width
			set flyme_width $eff_width

			set flyme_size [ expr $eff_width / 2 ]
		}

		set x [ expr $border_space + $flyme_width ]

		place $FM_R -x $x -y $y \
			-width $flyme_size -height $flyme_size

		set ffix [ expr 2 * $frame_border ]

		place $FM_I \
			-x [ expr $x + $frame_border ] \
			-y [ expr $y + $flyme_size + $frame_border ] \
			-width [ expr $flyme_size - $ffix ] \
			-height [ expr $flyme_size - $ffix ]

		set bs2 [ expr $border_space / 2 ]

		place $FM_I.fire_label -x $bs2 -y $bs2 -anchor nw

		update

		place $FM_IF -x [ right $FM_I.fire_label ] -y $bs2 -anchor nw \
			-width $row_height -height $row_height

		update

		set shy [ below $FM_IF ]

		place $FM_I.shield_label -x $bs2 -y $shy -anchor nw

		update

		set shx [ right $FM_I.shield_label ]

		set flyme_ssize [ expr $flyme_size - $shx - $border_space ]

		place $FM_IS -x $shx -y $shy \
			-width $flyme_ssize -height $row_height

		update

		set flyme_hsize [ expr $flyme_size - (4 * $row_height) \
			- (2 * $border_space) ]

		set hy [ expr $shy + $row_height ]

		place $FM_IH -x $bs2 -y $hy \
			-width $flyme_hsize -height $flyme_hsize

		update

		set hs2 [ expr $flyme_hsize / 2 ]

		set hi [ expr $flyme_hsize - (2 * $frame_border) ]
		set lo [ expr 2 * $frame_border ]

		$FM_IH coords $FM_IH_N $hs2 $lo
		$FM_IH coords $FM_IH_S $hs2 $hi
		$FM_IH coords $FM_IH_E $hi $hs2
		$FM_IH coords $FM_IH_W $lo $hs2

		set sy [ expr $hy + $flyme_hsize ]

		place $FM_I.speed -x $bs2 -y $sy

		update

		set sy [ expr $sy + $row_height ]

		place $FM_I.score -x $bs2 -y $sy
	} \
\
	else \
	{
		set flyme_height $height_left
		set flyme_width $width_left

		place forget $FM_R

		place forget $FM_I
	}

	place $FM_C -x $border_space -y $y \
		-width $flyme_width -height $flyme_height

	update

	#debug_proc layout_main_panel exit
}

