#
# $Id: slicer.tcl,v 1.25 1999/06/17 20:16:26 kohl Exp $
#

#
#                       CUMULVS version 1.0:
#       Computational Steering and Interactive Visualization
#                    in Distributed Applications
#            Oak Ridge National Laboratory, Oak Ridge TN.
#       Authors:  James Arthur Kohl and Philip M. Papadopoulos
#                    (C) 1996 All Rights Reserved
#
#                               NOTICE
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted
#  provided that the above copyright notice appear in all copies and
#  that both the copyright notice and this permission notice appear
#  in supporting documentation.
#
#  Neither the Institution, Oak Ridge National Laboratory, nor the
#  Authors make any representations about the suitability of this
#  software for any purpose.  This software is provided ``as is''
#  without express or implied warranty.
#
#  CUMULVS was funded by the U.S. Department of Energy.
#

#
# Main TCL Source File
#

#
# Hide Main Window
#

wm withdraw .

#
# Setup Main TCL Directories
#

set ckenv [ info exists env(STV_ROOT) ]

if { $ckenv == 1 } \
	{ set slicer_directory "$env(STV_ROOT)/examples/slicer" } \
\
else \
{
	set home	[ glob ~ ]
	set def		[ get_slicer_default_dir ]

	set slicer_directory "$home/$def"
}

set help_dir $slicer_directory/help

#
# Source Main TCL Procs
#

set ckdir_procs [ file exists procs.tcl ]

if { $ckdir_procs == 1 } \
	{ source procs.tcl } \
\
else \
	{ source $slicer_directory/procs.tcl }

#
# Source TCL Utility Procs
#

set ckdir_util [ file exists util.tcl ]

if { $ckdir_util == 1 } \
	{ source util.tcl } \
\
else \
	{ source $slicer_directory/util.tcl }

#
# Startup Message
#

puts ""

puts -nonewline "Initializing Slicer..."
flush stdout

#
# Proc Debug Flag
#

set proc_debug FALSE
#set proc_debug TRUE

#
# Create Toplevel Window
#

set main_height 750
set main_width 650

toplevel .slicer

wm withdraw .slicer

set_geometry -1 -1 125 50

wm minsize .slicer 530 400

set depth [ winfo depth .slicer ]

#
# Set up Colors
#

puts -nonewline "."
flush stdout

if { $depth == 1 } \
{
	set active_fg_color white

	set selector_color black

	set fg_color black

	set bg_color white
} \
\
else \
{
	set active_fg_color purple

	set selector_color purple

	set fg_color blue

	set bg_color #d9d9d9
}

#
# Reset Main Window Background
#

.slicer configure -background $bg_color

#
# Set Default Data Font (Fixed Width)
#

set data_font "-*-courier-bold-r-normal--14-*-*-*-*-*-*-*"

#
# Load Bitmaps
#

puts -nonewline "."
flush stdout

set xbm_dir "$slicer_directory/xbm"

# load_bitmap_file "$xbm_dir/right.xbm"

#
# Create Dummy Frame for Text Sizing (see get_real_text_size{})
#

puts -nonewline "."
flush stdout

frame .slicer.dummy

get_base_text_size

#
# Set Up Other Spacing Constants
#

set scroll_width 20

set border_space 10

set frame_border 2

#
# Help Window Constants
#

set min_help_height [ expr (3 * $scroll_width) + (2 * $border_space) ]

set min_help_width 250

set help_stagger_index 0

#
# Create Title & Version Number
#

puts -nonewline "."
flush stdout

set title_str "C U M U L V S    S l i c e r   [ title_info ]"

label .slicer.title -text $title_str \
	-foreground $fg_color -background $bg_color

place .slicer.title -relx 0.5 -y $border_space -anchor n

update

#
# Get Main Font for Entry Widgets
#

puts -nonewline "."
flush stdout

set main_font [ .slicer.title cget -font ]

#
# Create Main Panel
#

set SL .slicer

set SL_C $SL.canvas

set SL_M $SL.map

label $SL.message -text {Status:   Welcome to CUMULVS Slicer} \
	-foreground $fg_color -background $bg_color \
	-font $main_font -relief sunken -anchor nw

label $SL.help_msg -text {Help:} \
	-foreground $fg_color -background $bg_color \
	-font $main_font -relief sunken -anchor nw

set msg_list ""

#
# Main Menu
#

set file_menu_list \
	[ list \
		[ list "Quit Slicer"	command		quit \
			"Disconnect from Application and Exit Slicer" ] \
		[ list "Set Current"	command		set_current \
			"Set Current Field for Differential Comparison" ] \
		[ list "Free Current"	command		free_current \
			"Free Currently Set Field" ] \
		[ list "Save to File"	command		save_current \
			"Save Current Field to a File" ] \
		[ list "Load from File"	command		load_current \
			"Load Current Field from a File" ] \
	]

makeMenu $SL.file_menu fileHandle $file_menu_list lower {} none

button $SL.file -text "File..." -padx 1 -pady 1 \
	-font $main_font \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command [ list raiseMenu $SL.file_menu $SL.file $SL {} ]

restrict_bindings $SL.file ""

#
# Visualization Modes Menu
#

set mode_menu_list \
	[ list \
		[ list "Visualization Modes:"	label	vismodes ] \
		[ list "Normal Mode"	radiobutton		normal		vis_mode \
			"Set Visualization Mode to Normal" ] \
		[ list "Differential"	radiobutton		diff		vis_mode \
			"Set Visualization Mode to Differential" ] \
		[ list "Diff to Saved"	radiobutton		diffsave	vis_mode \
			"Set Visualization Mode to Difference with Saved" ] \
		[ list "Other Options:        "	label	optmodes ] \
		[ list "Auto Color Map"		statebutton		auto_color	OFF \
			state_auto_color \
			"Turn On / Off the Auto Color Map Ranging" ] \
		[ list "Show Min & Max"		statebutton		min_max		OFF \
			state_min_max \
			"Turn On / Off the Display of Data Value Min & Max" ] \
		[ list "Show Cumulative"	statebutton		cumulative	OFF \
			state_cumulative \
		"Turn On / Off the Display of Cumulative Min & Max Stats" ] \
		[ list "Reset Cumulative"	command			reset_cumulative \
			"Reset Cumulative Min & Max Stats" ] \
	]

set vis_mode "normal"

makeMenu $SL.mode_menu modeHandle $mode_menu_list lower {} none

button $SL.modes -text "Modes..." -padx 1 -pady 1 \
	-font $main_font \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color \
	-command [ list raiseMenu $SL.mode_menu $SL.modes $SL {} ]

restrict_bindings $SL.modes ""

#
# Field Selection
#

button $SL.field -text "Fields..." \
	-padx 1 -pady 1 \
	-foreground $fg_color -activeforeground $active_fg_color \
	-command "do_field"

restrict_bindings $SL.field ""

#
# Sleep Button
#

checkbutton $SL.sleep -text "Sleep" -command "sleepHandle" \
	-font $main_font -padx 1 -pady 1 \
	-bd $frame_border -relief raised \
	-onvalue ON -offvalue OFF -variable sleeping \
	-selectcolor $selector_color -foreground $fg_color \
	-activeforeground $active_fg_color

restrict_bindings $SL.sleep ""

set sleeping OFF

#
# Step Button
#

button $SL.step -text "Step Frame" \
	-padx 1 -pady 1 \
	-foreground $fg_color -activeforeground $active_fg_color \
	-command "do_step"

restrict_bindings $SL.step ""

#
# Lock Button
#

set lock_status "unlocked"

button $SL.lock -text "Lock Appl" \
	-padx 1 -pady 1 \
	-foreground $fg_color -activeforeground $active_fg_color \
	-command "do_lock toggle"

restrict_bindings $SL.lock ""

#
# Application Selection
#

label $SL.app -text "Application Name:" \
	-foreground $fg_color -background $bg_color

entry $SL.app_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.app_entry "Entry"

entry_setup $SL.app_entry "appHandle"

$SL.app_entry insert 0 [ get_app_name ]

#
# Plane Selection
#

label $SL.plane_label -text {Plane Selection:} \
	-foreground $fg_color -background $bg_color \
	-font $main_font -relief flat -anchor nw

radiobutton $SL.xy_plane -text { X-Y Plane } \
	-padx 1 -pady 1 \
	-command "handlePlane" \
	-bd $frame_border -relief sunken -value "XY" \
	-variable plane_choice \
	-selectcolor $selector_color \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color

restrict_bindings $SL.xy_plane ""

radiobutton $SL.yz_plane -text { Y-Z Plane } \
	-padx 1 -pady 1 \
	-command "handlePlane" \
	-bd $frame_border -relief sunken -value "YZ" \
	-variable plane_choice \
	-selectcolor $selector_color \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color

restrict_bindings $SL.yz_plane ""

radiobutton $SL.xz_plane -text { X-Z Plane } \
	-padx 1 -pady 1 \
	-command "handlePlane" \
	-bd $frame_border -relief sunken -value "XZ" \
	-variable plane_choice \
	-selectcolor $selector_color \
	-foreground $fg_color -background $bg_color \
	-activeforeground $active_fg_color

restrict_bindings $SL.xz_plane ""

set plane_choice "XY"

#
# Vis Frame Frequency
#

label $SL.freq -text "Freq:" \
	-foreground $fg_color -background $bg_color

entry $SL.freq_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.freq_entry "Entry"

entry_setup $SL.freq_entry "vis_freq_handle ; focus $SL.freq_entry"

$SL.freq_entry insert 0 1

#
# Vis Region Bounds & Cell
#

# "X" Axis

label $SL.xmin -text "X Axis Minimum:" \
	-foreground $fg_color -background $bg_color

entry $SL.xmin_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.xmin_entry "Entry"

entry_setup $SL.xmin_entry \
	"vis_region_handle $SL.xmin_entry $SL.xmax_entry"

$SL.xmin_entry insert 0 0

label $SL.xmax -text "X Axis Maximum:" \
	-foreground $fg_color -background $bg_color

entry $SL.xmax_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.xmax_entry "Entry"

entry_setup $SL.xmax_entry \
	"vis_region_handle $SL.xmax_entry $SL.xcell_entry"

$SL.xmax_entry insert 0 0

label $SL.xcell -text "X Axis Cell Size:" \
	-foreground $fg_color -background $bg_color

entry $SL.xcell_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.xcell_entry "Entry"

entry_setup $SL.xcell_entry \
	"vis_region_handle $SL.xcell_entry $SL.ymin_entry"

$SL.xcell_entry insert 0 1

# "Y" Axis

label $SL.ymin -text "Y Axis Minimum:" \
	-foreground $fg_color -background $bg_color

entry $SL.ymin_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.ymin_entry "Entry"

entry_setup $SL.ymin_entry \
	"vis_region_handle $SL.ymin_entry $SL.ymax_entry"

$SL.ymin_entry insert 0 0

label $SL.ymax -text "Y Axis Maximum:" \
	-foreground $fg_color -background $bg_color

entry $SL.ymax_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.ymax_entry "Entry"

entry_setup $SL.ymax_entry \
	"vis_region_handle $SL.ymax_entry $SL.ycell_entry"

$SL.ymax_entry insert 0 0

label $SL.ycell -text "Y Axis Cell Size:" \
	-foreground $fg_color -background $bg_color

entry $SL.ycell_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.ycell_entry "Entry"

entry_setup $SL.ycell_entry \
	"vis_region_handle $SL.ycell_entry $SL.zval_entry"

$SL.ycell_entry insert 0 1

# "Z" Axis

label $SL.zval -text "Z Axis Value:" \
	-foreground $fg_color -background $bg_color

entry $SL.zval_entry -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.zval_entry "Entry"

entry_setup $SL.zval_entry \
	"vis_region_handle $SL.zval_entry $SL.xmin_entry"

$SL.zval_entry insert 0 0

#
# Min & Max Stats
#

label $SL.minval -text "Min:" \
	-foreground $fg_color -background $bg_color

label $SL.maxval -text "Max:" \
	-foreground $fg_color -background $bg_color

#
# Color Map Canvas
#

canvas $SL_M -bd $frame_border -relief sunken -confine 1 \
	-background $bg_color

entry $SL.color_min -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.color_min "Entry"

entry_setup $SL.color_min "color_handle"

$SL.color_min insert 0 0

entry $SL.color_max -font $main_font \
	-bd $frame_border -relief sunken \
	-foreground $fg_color -background $bg_color

restrict_bindings $SL.color_max "Entry"

entry_setup $SL.color_max "color_handle"

$SL.color_max insert 0 100

#
# Main Canvas
#

canvas $SL_C -bd $frame_border -relief sunken -confine 1 \
	-background "#ffffff"
#	-background $bg_color

#
# Layout Main Panel
#

puts -nonewline "."
flush stdout

layout_main_panel

#
# Create Query Pop-Up
#

puts -nonewline "."
flush stdout

frame $SL.query -bd 4 -relief raised

restrict_bindings $SL.query "Frame"

label $SL.query.label \
	-text "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX" \
	-foreground $fg_color

place $SL.query.label -relx 0.50 -y [ expr $border_space / 2 ] -anchor n

update

place forget $SL.query

#
# Main Event Bindings
#

bind $SL.app_entry		<Enter>		"focus $SL.app_entry"
bind $SL.app_entry		<Leave>		"appHandle"

bind $SL.freq_entry		<Enter>		"focus $SL.freq_entry"
bind $SL.freq_entry		<Leave>		"vis_freq_handle"

bind $SL.xmin_entry		<Enter>		"focus $SL.xmin_entry"
bind $SL.xmin_entry		<Leave> \
	"vis_region_handle $SL.xmin_entry $SL.xmax_entry"

bind $SL.xmax_entry		<Enter>		"focus $SL.xmax_entry"
bind $SL.xmax_entry		<Leave> \
	"vis_region_handle $SL.xmax_entry $SL.xcell_entry"

bind $SL.xcell_entry	<Enter>		"focus $SL.xcell_entry"
bind $SL.xcell_entry	<Leave> \
	"vis_region_handle $SL.xcell_entry $SL.ymin_entry"

bind $SL.ymin_entry		<Enter>		"focus $SL.ymin_entry"
bind $SL.ymin_entry		<Leave> \
	"vis_region_handle $SL.ymin_entry $SL.ymax_entry"

bind $SL.ymax_entry		<Enter>		"focus $SL.ymax_entry"
bind $SL.ymax_entry		<Leave> \
	"vis_region_handle $SL.ymax_entry $SL.ycell_entry"

bind $SL.ycell_entry	<Enter>		"focus $SL.ycell_entry"
bind $SL.ycell_entry	<Leave> \
	"vis_region_handle $SL.ycell_entry $SL.zval_entry"

bind $SL.zval_entry		<Enter>		"focus $SL.zval_entry"
bind $SL.zval_entry		<Leave> \
	"vis_region_handle $SL.zval_entry $SL.xmin_entry"

bind $SL.color_min		<Enter>		"focus $SL.color_min"
bind $SL.color_min		<Leave>		"color_handle"

bind $SL.color_max		<Enter>		"focus $SL.color_max"
bind $SL.color_max		<Leave>		"color_handle"

bind $SL_C <ButtonPress-1>			"canvas_handle query_press %x %y"
bind $SL_C <Button1-Motion>			"canvas_handle query_slide %x %y"
bind $SL_C <ButtonRelease-1>		"canvas_handle query_release %x %y"

bind $SL_C <ButtonPress-2>			"canvas_handle zoom_press %x %y"
bind $SL_C <Button2-Motion>			"canvas_handle zoom_slide %x %y"
bind $SL_C <ButtonRelease-2>		"canvas_handle zoom_release %x %y"

bind $SL_C <3>						"canvas_handle pop %x %y"

set canvas_help \
	{Mouse Buttons: Left=Query, Middle_Sweep=Zoom_Area, Right=Unzoom.}

set canvas_help_cmd [ list setTmpMsg Help $canvas_help ]

bind $SL_C <Enter>					"focus $SL_C; \
										canvas_handle enter %x %y; \
										$canvas_help_cmd"

bind $SL_C <Motion>					"focus $SL_C; \
										canvas_handle move %x %y"

bind $SL_C <Leave>					"canvas_handle leave %x %y; popMsgs"

bind $SL_C <KeyPress>				"canvas_handle %A %x %y"

bind $SL <Configure>				"resize_main_panel"

bind $SL <Destroy>					"quit"

#
# Button Help Bindings
#

puts -nonewline "."
flush stdout

# main panel buttons

butt_help $SL.message label "Slicer Status Messages..."

butt_help $SL.help_msg label \
	"This is where Help Messages are Displayed..."

butt_help $SL.file button \
	"Raise / Lower Main File Menu - to Exit Slicer or Set Saved Fields."

butt_help $SL.modes button \
	"Raise / Lower Modes Menu - to Set Normal or Differential Viewing."

butt_help $SL.field button \
	"Raise / Lower Fields Menu - to Select from Available Data Fields."

butt_help $SL.sleep button \
	"Toggle Sleep State - to Pause / Unpause Any Connected Application."

butt_help $SL.step button \
	"Get Single Next Data Frame From Application (from Sleep State)."

butt_help $SL.lock button \
	"Lock / Unlock Application - Stall Execution for Fast Viz Response."

# application name

butt_help $SL.app label \
	"Name of the Application Currently Connected to Slicer."

butt_help $SL.app_entry entry \
	[ list "appHandle" \
	"Enter the Name of the Application to be Attached to by Slicer." ]

# slice plane selection

butt_help $SL.plane_label label \
	"Selection of the 2-D Plane to be Sliced Along for Viewing."

butt_help $SL.xy_plane button \
	"Select the X-Y Plane as the Slice Plane for Viewing."

butt_help $SL.yz_plane button \
	"Select the Y-Z Plane as the Slice Plane for Viewing."

butt_help $SL.xz_plane button \
	"Select the X-Z Plane as the Slice Plane for Viewing."

butt_help $SL.freq label \
	"Current Frequency of Data Frame Collection for Viewing."

butt_help $SL.freq_entry entry \
	[ list "vis_freq_handle ; focus $SL.freq_entry" \
	"Enter the Desired Data Frame Collection Frequency for Viewing." ]

# primary axis range

butt_help $SL.xmin label \
	"Minimum Element Index for the Primary Slice Axis."

butt_help $SL.xmin_entry entry \
	[ list "vis_region_handle $SL.xmin_entry $SL.xmax_entry" \
	"Enter the Minimum Element Index for the Primary Slice Axis." ]

butt_help $SL.xmax label \
	"Maximum Element Index for the Primary Slice Axis."

butt_help $SL.xmax_entry entry \
	[ list "vis_region_handle $SL.xmax_entry $SL.xcell_entry" \
	"Enter the Maximum Element Index for the Primary Slice Axis." ]

butt_help $SL.xcell label \
	"Cell Size (Resolution) of Data Samples for Primary Slice Axis."

butt_help $SL.xcell_entry entry \
	[ list "vis_region_handle $SL.xcell_entry $SL.ymin_entry" \
	"Enter the Data Sample Cell Size for the Primary Slice Axis." ]

# secondary axis range

butt_help $SL.ymin label \
	"Minimum Element Index for the Secondary Slice Axis."

butt_help $SL.ymin_entry entry \
	[ list "vis_region_handle $SL.ymin_entry $SL.ymax_entry" \
	"Enter the Minimum Element Index for the Secondary Slice Axis." ]

butt_help $SL.ymax label \
	"Maximum Element Index for the Secondary Slice Axis."

butt_help $SL.ymax_entry entry \
	[ list "vis_region_handle $SL.ymax_entry $SL.ycell_entry" \
	"Enter the Maximum Element Index for the Secondary Slice Axis." ]

butt_help $SL.ycell label \
	"Cell Size (Resolution) of Data Samples for Secondary Slice Axis."

butt_help $SL.ycell_entry entry \
	[ list "vis_region_handle $SL.ycell_entry $SL.zval_entry" \
	"Enter the Data Sample Cell Size for the Secondary Slice Axis." ]

# slice plane index

butt_help $SL.zval label \
	"Element Index for the Location of the Slice Plane."

butt_help $SL.zval_entry entry \
	[ list "vis_region_handle $SL.zval_entry $SL.xmin_entry" \
	"Enter the Element Index for the Location of the Slice Plane." ]

# color map

butt_help $SL.color_min entry \
	[ list "color_handle" \
	"Enter the Minimum Data Value for the Color Mapping Range." ]

butt_help $SL.color_max entry \
	[ list "color_handle" \
	"Enter the Maximum Data Value for the Color Mapping Range." ]

butt_help $SL_M label \
	"Color Map for Data Value Range."

#
# Actually Bring Up Main XPVM Window
#

wm deiconify .slicer

#
# Reset Welcome Message
#

puts "done."
flush stdout

setMsg "Welcome to CUMULVS Slicer"

update

