
/* $Id: stv.h,v 3.90 1999/11/16 00:02:31 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


#ifndef _STV_H_

#define _STV_H_


/* Defines for CUMULVS steering, visualization and checkpointing */


/* Debugging Flags */

/* #define STV_MEM_DEBUG */

#define DEBUG

#define STV_DEBUG_MSG			0x00001
#define STV_DEBUG_SPAWN			0x00002
#define STV_DEBUG_ULIB			0x00004
#define STV_DEBUG_FLUSH			0x00008
#define STV_DEBUG_NOTIFY		0x00010
#define STV_DEBUG_MEM			0x00020
#define STV_DEBUG_TIME			0x00040
#define STV_DEBUG_FREQ			0x00080
#define STV_DEBUG_PACK			0x00100
#define STV_DEBUG_PACK_INDEX	0x00200
#define STV_DEBUG_VIEWER		0x00400
#define STV_DEBUG_STEER			0x00800
#define STV_DEBUG_APPINFO		0x01000
#define STV_DEBUG_CPD			0x02000
#define STV_DEBUG_RESTART		0x04000
#define STV_DEBUG_REGION		0x08000
#define STV_DEBUG_CKPT			0x10000
#define STV_DEBUG_CKPT_DATA		0x20000

#define STV_DEBUG_ALL			0x3ffff


/* Group Name Constants */

#define stvDefaultClient		"defaultClient" 
#define stvPrefix				"stv_"
#define stvCpdGroup				"stv__cpd"
#define stvCpdWorkGroup			"stv__cpdwork"

#define stvMaxNameLen			255


/* Mbox Names */

#define stvViewerContext		"stv__VIEWER__CTXT"
#define stvCpdContext			"stv__CPD__CTXT"
#define stvAppContext			"stv__APP__CTXT"


/* -----------  Message types  ----------- */

#define stvMsgOp( _const )		( -1 * (_const) )

#define stvNotifyOp( _op )		( _op < 0 )

#define stvInit0						100
#define stvInit1						101

#define stvFieldRqst					200
#define stvFieldTime					201
#define stvFieldTstart					202
#define stvFieldAck						203
#define stvFieldHaltGroup				204
#define stvFieldHaltAll					205

#define stvXon							300
#define stvRegion						301
#define stvVisRegion					302
#define stvVisFrequency					303
#define stvNewParams					304
#define stvTimeSync						305
#define stvAppLock						306
#define stvAppUnlock					307
#define stvSetopt						308
#define stvSteerRelease					309

#define stvNotify						400
#define stvReconfigure					401
#define stvDisconnect					402

#define stvCpdAout						500
#define stvCpdUserNotifyInfo			501
#define stvCheckpointDataSubmit			502
#define stvCheckpointDataUpdate			503
#define stvCheckpointForward			504
#define stvCpdXon						505
#define stvConfigCpd					506
#define stvRequestWhole					507
#define stvRequestCheckpoint			508
#define stvWholeCheckpoint				509
#define stvRestart						510
#define stvCpdSpawn						511
#define stvCpdStabilizeRqst				512
#define stvCpdStabilizeReply			513
#define stvCpdCheckpointStable			514
#define stvCpdVerifyCheckpointRelease	515
#define stvCheckpointRelease			516
#define stvCpdNproc						517
#define stvCpdNewPeer					518
#define stvCpdPeerFailed				519
#define stvCpdForwardPeerFailed			520
#define stvAppFinished					521
#define stvCpdVerifyAppFinished			522
#define stvAppRelease					523
#define stvCpdListFiles					524
#define stvCpdFileList					525
#define stvCpdReconfigRqst				526
#define stvCpdReconfigReply				527
#define stvCpdBootstrap					528
#define stvCpdHalt						529
#define stvCpdShutdown					530

#define stvSenderFailed			stvFieldHaltAll


/* -----------  Message Markers  ----------- */

#define stvMarkerParam				-1
#define stvMarkerParamEnd			-2
#define stvMarkerDecomp				-3
#define stvMarkerDecompEnd			-4
#define stvMarkerParticle			-5
#define stvMarkerParticleEnd		-6
#define stvMarkerField				-7
#define stvMarkerFieldDecomp		-8
#define stvMarkerFieldParticle		-9
#define stvMarkerFieldNone			-10
#define stvMarkerFieldData			-11
#define stvMarkerFieldDataEnd		-12
#define stvMarkerFieldEnd			-13
#define stvMarkerParticleField		-14
#define stvMarkerParticleFieldEnd	-15
#define stvMarkerSubRegion			-16
#define stvMarkerSubRegionEnd		-17
#define stvMarkerCp					-18
#define stvMarkerCpEnd				-19
#define stvMarkerParamValues		-20
#define stvMarkerParamValuesEnd		-21


/* --------  Message Receive Types -------- */

#define stvMsgNoBlock				0
#define stvMsgBlock					1
#define stvMsgBlockTimeout			2
#define stvMsgProbe					3
#define stvMsgProbeBlock			4


/* ---------- Parameter Use Flags ---------- */

#define stvVisOnly					1
#define stvCpOnly					2

#define stvVisCp					( stvVisOnly + stvCpOnly )


/* ------------ Steering States ------- */

#define stvSteerNot					0
#define stvSteerRqstd				1
#define stvSteerToken				2


/* ------ Application Lock States ----- */

#define stvAppUnlocked				0
#define stvAppLocked				1


/* ------------ Return Codes ---------- */

#define stvStatusOk					 0
#define stvStatusBadFrame			-1
#define stvStatusIncomplete			-2
#define stvStatusNoMem				-3
#define stvStatusBadParam			-4
#define stvStatusSysFail			-5
#define stvStatusBail				-6
#define stvStatusReconfigure		-7
#define stvStatusDisconnect			-8
#define stvStatusCpdDupServer		-9
#define stvStatusMax				-10


/* --- Field Request Flag Constants --- */

#define stvFieldRqstDefault			0
#define stvFieldRqstNoSynch			1
#define stvFieldRqstStatic			2
#define stvFieldRqstNoSynchFrame	4


/* ------- Time Scale Constants ------ */

/* Bit Field Constructed as Follows: */
/* stvTimeBase is:  0100 0000 0000 0000 0000 0000 0000 0000    */
/* time unit values reside in 5 bits: 27 - 31, w/bit 32 = sign */

#define stvTimeNegative		( (unsigned) 1 \
								<< ( (int) ( (8 * sizeof(int)) - 1 ) ) )

#define stvTimeBase			( (unsigned) 1 \
								<< ( (int) ( (8 * sizeof(int)) - 2 ) ) )

#define stvTimeShift		( (int) ( (8 * sizeof(int)) - 6 ) )

/* Shift 4-bit Value to Butt Up Against stvTimeBase... */

#define stvTimeMaxValue		( ( (unsigned) 1 << stvTimeShift ) - 1 )

#define stvTimePicoSec		( stvTimeBase )
#define stvTimeNanoSec		( stvTimeBase + ( 0x1 << stvTimeShift ) )
#define stvTimeMicroSec		( stvTimeBase + ( 0x2 << stvTimeShift ) )
#define stvTimeMilliSec		( stvTimeBase + ( 0x3 << stvTimeShift ) )
#define stvTimeSec			( stvTimeBase + ( 0x4 << stvTimeShift ) )
#define stvTimeMin			( stvTimeBase + ( 0x5 << stvTimeShift ) )
#define stvTimeHour			( stvTimeBase + ( 0x6 << stvTimeShift ) )
#define stvTimeDay			( stvTimeBase + ( 0x7 << stvTimeShift ) )
#define stvTimeWeek			( stvTimeBase + ( 0x8 << stvTimeShift ) )
#define stvTimeYear			( stvTimeBase + ( 0x9 << stvTimeShift ) )
#define stvTimeDecade		( stvTimeBase + ( 0xa << stvTimeShift ) )
#define stvTimeCentury		( stvTimeBase + ( 0xb << stvTimeShift ) )
/* Note:  after 0xf, need to add another bit! */

#define stvTimeNone			0

#define stv_time_of( _value, _units ) \
	( ( (_value) >= 0 ) ? \
		( ( (unsigned) (_value) > (unsigned) stvTimeMaxValue ) ? \
			stv_normalize_time_of( _value, _units ) \
			: ( (_units) + (_value) ) \
		) \
		: ( stvTimeNegative \
			| ( ( (unsigned) ( -1 * (_value) ) \
					> (unsigned) stvTimeMaxValue ) ? \
				stv_normalize_time_of( -1 * (_value), _units ) \
				: ( (_units) + ( -1 * (_value) ) ) \
			) \
		) \
	)

/* It's a sign of the times...  la la la la... */
#define stv_sign_of_time( _tv ) \
	( ( (_tv) & stvTimeNegative ) ? -1 : 1 )

#define stv_value_of_time( _tv ) \
	( stv_sign_of_time( _tv ) * ( (_tv) & stvTimeMaxValue ) )

#define stv_unit_of_time( _tv ) \
	( (_tv) & ( (~stvTimeMaxValue) & (stvTimeNegative - 1) ) )

#define stv_is_time_value( _tv ) \
	( (_tv) & stvTimeBase )

#define stvNotTime	(~stvTimeBase)


/* ---------- Misc Constants --------- */

#define stvDefault					-1

#define stvAny						-1

#define stvFalse					0
#define stvTrue						1

#define stvTerse					0
#define stvVerbose					1

#define stvSystemInitd				1
#define stvSystemActive				2

#define stvCpdTag					100

#define stvCpdHoleTypeUnpack		1
#define stvCpdHoleTypePack			2
#define stvCpdHoleTypeWrite			3
#define stvCpdHoleTypeRead			4

#define stvCpdHoleDataInt			1
#define stvCpdHoleDataLong			2
#define stvCpdHoleDataULong			3
#define stvCpdHoleDataStr			4
#define stvCpdHoleDataData			5

#define stvDefaultInfo				( (int *) NULL )
#define stvLocalArray				( (int *) NULL )

#define stvNoFieldOffsets			( (int *) NULL )


/* ----------- Checkpoint Daemon, Application States ------- */

#define stvCpdStable				1
#define stvCpdAppFail				2
#define stvCpdAppReconfig			3
#define stvCpdAppStable				4
#define stvCpdNewApp				5
#define stvCpdWholeRequested		7
#define stvCpdRestart				8
#define stvCpdAppFlux				9
#define stvCpdReplyRecvd			10
#define stvCpdCheckpointStabilizing	11
#define stvCpdWaitStabilize			12
#define stvCpdIsStable				13
#define stvCpdWaitStabEpoch			14
#define stvCpdGotStabEpoch			15
#define stvCpdWaitReconfig			16
#define stvCpdGotReconfig			17


/* ------------ Data Types ----------- */

#define stvStr						0		/* string */
#define stvByte						1		/* byte */
#define stvShort					2		/* short */
#define stvInt						3		/* int */
#define stvFloat					4		/* real */
#define stvCplx						5		/* complex */
#define stvDouble					6		/* double */
#define stvDcplx					7		/* double complex */
#define stvLong						8		/* long integer */
#define stvUshort					9		/* unsigned short int */
#define stvUint						10		/* unsigned int */
#define stvUlong					11		/* unsigned long int */

#define stvTypeMask					0x1f	/* data type, w/o index */


/* --------------- Parameter Index Constant -------------- */
/* -- must be bit-wise orthogonal with above data types -- */

#define stvIndex					32


/* -------- Data Storage Constants --------- */

#define stvColumnMajor				0
#define stvRowMajor					1


/* ------------ Connection States --------- */

#define stvNotConnected				0
#define stvConnected				1
#define stvWaitTstart				2
#define stvConnectDone				3
#define stvDeadConnection			4


/* --- stv_sendReadyData() flag constants --- */

#define stvSendDefault				0
#define stvSendNoFieldTimeUpdates	1


/* ----------- Option IDs ------------------ */

#define stvOptAll					-2
#define stvOptNull					-1
#define stvOptFullUnpack			0
#define stvOptVerbose				1
#define stvOptDumpAll				2
#define stvOptDump					3
#define stvOptFieldStorage			4
#define stvOptAppLock				5
#define stvOptCpRecovery			6
#define stvOptFieldTimeScale		7
#define stvOptSetFieldTime			8
#define stvOptIncrFieldTime			9
#define stvOptFieldReady			10
#define stvOptMax					11


/* ----------- Cp Option IDs -------------- */

#define stvOptCpRestart				1
#define stvOptCpRollback			2


/* ---------- Decomposition Types ---------- */

#define stvBlock					1
#define stvCyclic					2
#define stvExplicit					3
#define stvCollapse					4


/* Miscellaneous Constants */

#define STV_MAX_DECOMP				100
#define STV_MAX_PARAM				100
#define STV_MAX_FIELD				100

#define STV_MAX_DIM					3


/* ------------- Macros ------------- */

#define STV_GLOBAL_CYCLE_INDEX( _ci, _cnum, _csz, _glb, _p, _ntasks ) \
	( ( (_p) * (_csz) ) \
		+ ( (_cnum) * (_csz) * (_ntasks) ) + (_ci) \
		+ (_glb) )

#define STV_LOCAL_CYCLE_INDEX( _ci, _cnum, _csz ) \
	( ( (_cnum) * (_csz) ) + (_ci) )

#define STV_START_BLOCK_INDEX( _index, _offset, _cellsize ) \
	( (_index) + \
		( ( ((_index) - (_offset)) % (_cellsize) ) == 0 ? \
			0 : \
			( (_cellsize) \
				- ( ((_index) - (_offset)) % (_cellsize) \
					+ ( ((_index) < (_offset)) ? (_cellsize) : 0 ) ) \
			) \
		) \
	)


#define STV_DIVISIBLE( _dvd, _dvr ) \
	( ( ( (_dvd) / (_dvr) ) * (_dvr) ) == (_dvd) )


#define STV_VALUE_OF( _ptr, _type )    ( *( (_type *) _ptr ) )

#define STV_PTR_OF( _ptr, _type )    ( (_type *) _ptr )

#define STV_SIZE_T	size_t


#ifndef MAX
#define MAX( _x, _y )    ( (_x) > (_y) ? (_x) : (_y) )
#endif

#ifndef MIN
#define MIN( _x, _y )    ( (_x) < (_y) ? (_x) : (_y) )
#endif


#define STV_EQUIV_DELTA			0.000001

#define STV_EQUIV( _d1, _d2 ) \
( ( (_d1) != 0.0 ) ? \
	( (stv_fabs((_d1) - (_d2)) / stv_fabs( _d1 )) < STV_EQUIV_DELTA ) \
	: ( stv_fabs( (_d1) - (_d2) ) < STV_EQUIV_DELTA ) )


/* Library Routine Prototype Macro */

#ifdef __STV_Prototype__
#undef __STV_Prototype__
#endif

#if defined(__STDC__) || defined(__cplusplus)
#define __STV_Prototype__(x) x
#else
#define __STV_Prototype__(x) ()
#endif

#ifdef __STDC__
#include <stdlib.h>
#endif

#ifdef  SYSVSTR
#include <string.h>
#else
#include <strings.h>
#endif

#ifndef WIN32
#include <sys/time.h>
#else
#include <winsock.h>
#endif


/* Other System Header Files */

#include <stv_mp.h>
#include <stv_type.h>
#include <stv_proto.h>


/* _STV_H_ */
#endif

