
/* $Id: stv_glob.c,v 3.28 2000/07/18 14:05:29 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>


/* STV Header Files */

#include "stv.h"


/* STV Globals */


/* System Globals */

STV_VIEW_FIELD_GROUP STV_STEER_VFG = (STV_VIEW_FIELD_GROUP) NULL;

STV_VIEW_FIELD_GROUP STV_SYNC_VFG = (STV_VIEW_FIELD_GROUP) NULL;

STV_CONTEXT STV_APP_CTX = (STV_CONTEXT) -1;
STV_CONTEXT STV_CPD_CTX = (STV_CONTEXT) -1;
STV_CONTEXT STV_VIEWER_CTX = (STV_CONTEXT) -1;

struct timeval STV_TIMEOUT = { 10, 0 };

char STVUSERAPPNAME[256];
char STVSHADOWNAME[256];
char STVAPPNAME[256];

int STV_DUMMY_AXIS_INFO[3] = { stvDefault, stvDefault, stvDefault };
int STV_DUMMY_AXIS_TYPE[3] = { stvCollapse, stvCollapse, stvCollapse };

int STV_APP_LOCK = stvAppUnlocked;
int STV_APP_LOCK_HOLD = -1;
int STV_APP_LOCK_TIME = -1;

int STV_DEACTIVATED = -1;

int STV_FULLGROUP = 0;

int STV_GCOORD = -1;

int STV_MYNODEID = -1;
int STV_MYID = -1;

int STV_NTASKS = 0;

int STV_TIMESTAMP = 0;

int STV_CP_RECOVERY_HOLD = stvFalse;    /* ignore Init0's until done! */

int STV_NRESTARTS = 0;    /* # of times has this task restarted? */

int STV_USERTAG = -1;     /* user tag for messages to/from avs */


/* Information Lists' and/or Static Allocation */

STV_VIEWER STVVIEWERS = (STV_VIEWER) NULL;
int nSTVVIEWERS = 0;

STV_BUFFER STVBUFFERS;
int nSTVBUFFERS = 0;

STV_HANDLER STVHANDLERS = (STV_HANDLER) NULL;
int nSTVHANDLERS = 0;

STV_HANDLER STVNOTIFYHANDLERS = (STV_HANDLER) NULL;
int nSTVNOTIFYHANDLERS = 0;

STV_MSG_PASS_THRU STVMSGPASSTHRUS = (STV_MSG_PASS_THRU) NULL;
int nSTVMSGPASSTHRUS = 0;

STV_FIELD STVFIELDS;
int nSTVFIELDS = 0;

STV_PARAM STVPARAMS;
int nSTVPARAMS = 0;

STV_PARTICLE STVPARTICLES;
int nSTVPARTICLES = 0;


/* ID Counter Globals */

int STV_VIEWER_ID = 0;

int STV_INIT_ID = 0;

int STV_NEXT_FGID = 0;

int STV_NEXT_REGION_ID = 0;


/* Option Globals */

int STV_OPT_CP_RECOVERY = stvOptCpRestart;

int STV_OPT_DUMP = stvFalse;

int STV_OPT_DUMP_ALL = stvFalse;

int STV_OPT_FULL_UNPACK = stvFalse;

int STV_OPT_VERBOSE = stvFalse;


/* Global Temporary Storage for Macros */

STV_CONTEXT STV_TMP_CTX;

void *STV_TMP_PTR;

int STV_TMP_CC;
int STV_TMP_CC2;


/* Checkpointing Globals */

CPD_DATAPTR CPDATA = ( CPD_DATAPTR ) NULL;

int CPDTID = -1;        /* tid of my checkpointing daemon */

int nCPFIELDS = 0;
int nCPPARAMS = 0;


/* Time Conversion Data */

TIME_CONVERSION STV_TIME_CONV[] =
{
	stvTimePicoSec,		1000,	stvTimeNanoSec,
	stvTimeNanoSec,		1000,	stvTimeMicroSec,
	stvTimeMicroSec,	1000,	stvTimeMilliSec,
	stvTimeMilliSec, 	1000,	stvTimeSec,
	stvTimeSec,			  60,	stvTimeMin,
	stvTimeMin,			  60,	stvTimeHour,
	stvTimeHour,		  24,	stvTimeDay,
	stvTimeDay,			   7,	stvTimeWeek,
	stvTimeWeek,		  52,	stvTimeYear,
	stvTimeYear,		  10,	stvTimeDecade,
	stvTimeDecade,		  10,	stvTimeCentury,
	0,					   0,	0
};


/* Error Text Storage */

char STVERRTXT[1024];

char *STVERRNAMES[ -(stvStatusMax) ] =
{
	"stvStatusOk",
	"stvStatusBadFrame",
	"stvStatusIncomplete",
	"stvStatusNoMem",
	"stvStatusBadParam",
	"stvStatusSysFail",
	"stvStatusBail",
	"stvStatusReconfigure",
	"stvStatusDisconnect",
	"stvStatusCpdDupServer"
};

