
/* $Id: stv_mp.c,v 3.4 1999/09/30 20:22:40 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"


/* Message passing support routines */


#ifdef STV_MP_PVM


int
stv_host_config( stv_hostp, nhostp )
STV_HOSTPTR *stv_hostp;
int *nhostp;
{
	struct pvmhostinfo *hostp = 0;

	int nhosts;
	int cc;
	int i;

	cc = pvm_config( &nhosts, (int *) NULL, &hostp );

	if ( cc >= 0 )
	{
		if ( stv_hostp != NULL )
		{
			*stv_hostp = (STV_HOSTPTR) MALLOC( (STV_SIZE_T) nhosts
				* sizeof( STV_HOST ) );
			if ( !(*stv_hostp) )
				return( stvFalse );

			for ( i=0 ; i < nhosts ; i++ )
			{
				strcpy( (*stv_hostp)[i].name, hostp[i].hi_name );
				strcpy( (*stv_hostp)[i].arch, hostp[i].hi_arch );
				(*stv_hostp)[i].tid = hostp[i].hi_tid;
			}
		}

		if ( nhostp != NULL )
			*nhostp = nhosts;

		return( stvTrue );
	}

	return( stvFalse );
}


/* Name Service Routines */

#if ( PVM_MAJOR_VERSION > 3 ) || ( PVM_MINOR_VERSION >= 4 )

/* Do Nothing */

#else

#ifdef CRAY

/* Spank the Cray Name Database...  :-) */

/* Also Necessary for T3E Cray: */
/* setenv PVM_GROUPS_FULL 1     */
/* setenv PVM_PE_LIST all       */

/*
 *         PVM version 3.3:  Parallel Virtual Machine System
 *               University of Tennessee, Knoxville TN.
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *                   Emory University, Atlanta GA.
 *      Authors:  A. L. Beguelin, J. J. Dongarra, G. A. Geist,
 *    W. C. Jiang, R. J. Manchek, B. K. Moore, and V. S. Sunderam
 *                   (C) 1992 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * Neither the Institutions (Emory University, Oak Ridge National
 * Laboratory, and University of Tennessee) nor the Authors make any
 * representations about the suitability of this software for any
 * purpose.  This software is provided ``as is'' without express or
 * implied warranty.
 *
 * PVM version 3 was funded in part by the U.S. Department of Energy,
 * the National Science Foundation and the State of Tennessee.
 */


/***************
 **  Globals  **
 ***************/

#define TMDB_INSERT     1
#define TMDB_DELETE     2
#define TMDB_LOOKUP     3

#define 	TIDPVMD		0x80000000
#define 	TM_DB 		((int)0x80010001) + 15 

#define BEATASK (pvmmytid == -1 ? ( pvmmytid = pvm_mytid(), 0) : 0)
#define TEV_DO_TRACE(_x) 0
#define TEV_FIN

static int pvmmytid = -1;

int
pvm_delete(name, req)
	char *name;		/* class name */
	int req;		/* class index or -1 for all */
{
	int sbf, rbf, cc;
	int x;

	if (!name || !*name || req < -1) {
		cc = PvmBadParam;

	} else {
		if (!(cc = BEATASK)) {
			sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
			rbf = pvm_setrbuf(0);
			cc = TMDB_DELETE;
			pvm_pkint(&cc, 1, 1);
			pvm_pkstr(name);
			pvm_pkint(&req, 1, 1);
			if ((cc = msendrecv(TIDPVMD, TM_DB)) > 0) {
				pvm_upkint(&cc, 1, 1);
				pvm_freebuf(pvm_setrbuf(rbf));

			} else
				pvm_setrbuf(rbf);
			pvm_freebuf(pvm_setsbuf(sbf));
		}
	}

	if (cc < 0 && cc != PvmNoEntry)
		lpvmerr("pvm_delete", cc);
	return cc;
}


int
pvm_insert(name, req, data)
	char *name;		/* class name */
	int req;		/* requested class index or -1 for any */
	int data;
{
	int sbf, rbf, cc;
	int x;

	if (!name || !*name || req < -1) {
		cc = PvmBadParam;

	} else {
		if (!(cc = BEATASK)) {
			sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
			rbf = pvm_setrbuf(0);
			cc = TMDB_INSERT;
			pvm_pkint(&cc, 1, 1);
			pvm_pkstr(name);
			pvm_pkint(&req, 1, 1);
			pvm_pkint(&data, 1, 1);
			if ((cc = msendrecv(TIDPVMD, TM_DB)) > 0) {
				pvm_upkint(&cc, 1, 1);
				pvm_freebuf(pvm_setrbuf(rbf));

			} else
				pvm_setrbuf(rbf);
			pvm_freebuf(pvm_setsbuf(sbf));
		}
	}

	if (cc < 0 && cc != PvmDupEntry)
		lpvmerr("pvm_insert", cc);
	return cc;
}


int
pvm_lookup(name, req, datap)
	char *name;		/* class name */
	int req;		/* req class index or -1 for any */
	int *datap;		/* data return */
{
	int sbf, rbf, cc, cc1;
	int x, nbytes;
	short wfcs;

	if (!name || !*name || req < -1) {
		cc = PvmBadParam;

	} else {
		if (!(cc = BEATASK)) {
			sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
			rbf = pvm_setrbuf(0);
			cc = TMDB_LOOKUP;
			pvm_pkint(&cc, 1, 1);
			pvm_pkstr(name);
			pvm_pkint(&req, 1, 1);
			if ((cc = msendrecv(TIDPVMD, TM_DB)) >= 0) {
				cc1 = pvm_upkshort(&wfcs, 1, 1);
				cc = wfcs;
				if (cc >= 0 && datap)
				{
					cc = pvm_upkshort(&wfcs, 1, 1);
					*datap = wfcs;
				}
				pvm_freebuf(pvm_setrbuf(rbf));

			} else
				pvm_setrbuf(rbf);
			pvm_freebuf(pvm_setsbuf(sbf));
		}
	}


	if (cc < 0 && cc != PvmNoEntry)
		lpvmerr("pvm_lookup", cc);
	return cc;
}

/* CRAY */
#endif


/* PVM 3.3 */
#endif


/* STV_MP_PVM */
#endif

