
/* $Id: stv_util.c,v 3.104 2000/07/14 19:16:03 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <malloc.h>
#include <math.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


/* CUMULVS Banner */

void
stv_banner()
{
	char *version;
	char *name;

	printf( "\n" );

	version = stv_version();
	name = stv_version_name();

	printf( "CUMULVS Version %s (%s)\n", version, name );

	free( version );

	printf( "Oak Ridge National Laboratory, Oak Ridge, TN\n" );
	printf( "(C) 1996-1998 All Rights Reserved\n" );
	printf( "CUMULVS was funded by the U.S. Department of Energy.\n" );

	printf( "\n" );
}


/* User Library & Viewer Options */

int
stv_setopt( what, option, value )
int what;
int option;
int value;
{
	STV_FIELDPTR F;

	int oldvalue;
	int tunits;
	int tval;

	switch ( option )
	{
		case stvOptFullUnpack:
		{
			oldvalue = STV_OPT_FULL_UNPACK;

			STV_OPT_FULL_UNPACK = value ? stvTrue : stvFalse;

			break;
		}

		case stvOptVerbose:
		{
			oldvalue = STV_OPT_VERBOSE;

			STV_OPT_VERBOSE = value;

			break;
		}

		case stvOptDumpAll:
		{
			oldvalue = STV_OPT_DUMP_ALL;

			STV_OPT_DUMP_ALL = value ? stvTrue : stvFalse;

			break;
		}

		case stvOptDump:
		{
			oldvalue = STV_OPT_DUMP;

			STV_OPT_DUMP = value ? stvTrue : stvFalse;

			break;
		}

		case stvOptAppLock:
		{
			oldvalue = STV_APP_LOCK;

			STV_APP_LOCK = value ? stvAppLocked : stvAppUnlocked;

			break;
		}

		case stvOptCpRecovery:
		{
			oldvalue = STV_OPT_CP_RECOVERY;

			STV_OPT_CP_RECOVERY = value;

			break;
		}

		case stvOptFieldStorage:
		{
			if ( value != stvRowMajor && value != stvColumnMajor )
			{
				sprintf( STVERRTXT, "Unknown Storage Order %d",
					value );

				STV_ERROR( "stv_setopt", stvStatusBadParam, STVERRTXT );

				break;
			}

			if ( what != stvOptAll )
			{
				if ( (F = (STV_FIELDPTR) stv_findField(
						&STVFIELDS, what )) == (STV_FIELDPTR) NULL )
				{
					sprintf( STVERRTXT, "Unknown Field ID %d", what );

					STV_ERROR( "stv_setopt", stvStatusBadParam,
							STVERRTXT );

					oldvalue = -1;

					break;
				}	

				oldvalue = F->storage_order;

				F->storage_order = value;
			}

			else
			{
				F = STVFIELDS.next;

				oldvalue = -1;

				while ( F != NULL )
				{
					/* return the last field's "old value"... */
					oldvalue = F->storage_order;

					F->storage_order = value;

					F = F->next;
				}
			}

			break;
		}

		case stvOptFieldTimeScale:
		{
			if ( !stv_is_time_value( value ) )
			{
				sprintf( STVERRTXT, "Invalid Time Value %d", value );

				STV_ERROR( "stv_setopt", stvStatusBadParam, STVERRTXT );

				break;
			}

			if ( what != stvOptAll )
			{
				if ( (F = (STV_FIELDPTR) stv_findField(
						&STVFIELDS, what )) == (STV_FIELDPTR) NULL )
				{
					sprintf( STVERRTXT, "Unknown Field ID %d", what );

					STV_ERROR( "stv_setopt", stvStatusBadParam,
							STVERRTXT );

					oldvalue = -1;

					break;
				}	

				oldvalue = F->time_scale;

				F->time_scale = value;

				/* If base time not set, munge the units... */
				if ( F->time_base == 0 && F->time_units == stvTimeNone )
					F->time_units = stv_unit_of_time( value );
			}

			else
			{
				F = STVFIELDS.next;

				oldvalue = -1;

				while ( F != NULL )
				{
					/* return the last field's "old value"... */
					oldvalue = F->time_scale;

					F->time_scale = value;

					/* If base time not set, munge the units... */
					if ( F->time_base == 0
							&& F->time_units == stvTimeNone )
						F->time_units = stv_unit_of_time( value );

					F = F->next;
				}
			}

			break;
		}

		case stvOptSetFieldTime:
		{
			if ( !stv_is_time_value( value ) )
			{
				sprintf( STVERRTXT, "Invalid Time Value %d", value );

				STV_ERROR( "stv_setopt", stvStatusBadParam, STVERRTXT );

				break;
			}

			if ( what != stvOptAll )
			{
				if ( (F = (STV_FIELDPTR) stv_findField(
						&STVFIELDS, what )) == (STV_FIELDPTR) NULL )
				{
					sprintf( STVERRTXT, "Unknown Field ID %d", what );

					STV_ERROR( "stv_setopt", stvStatusBadParam,
							STVERRTXT );

					oldvalue = -1;

					break;
				}	

				oldvalue = stv_time_of( F->time_base, F->time_units );

				F->time_base = stv_value_of_time( value );
				
				F->time_units = stv_unit_of_time( value );
			}

			else
			{
				F = STVFIELDS.next;

				oldvalue = -1;

				while ( F != NULL )
				{
					/* return the last field's "old value"... */
					oldvalue = stv_time_of( F->time_base,
							F->time_units );

					F->time_base = stv_value_of_time( value );
				
					F->time_units = stv_unit_of_time( value );

					F = F->next;
				}
			}

			break;
		}

		case stvOptIncrFieldTime:
		{
			if ( !stv_is_time_value( value ) )
			{
				sprintf( STVERRTXT, "Invalid Time Value %d", value );

				STV_ERROR( "stv_setopt", stvStatusBadParam, STVERRTXT );

				break;
			}

			tval = stv_value_of_time( value );
			tunits = stv_unit_of_time( value );

			if ( what != stvOptAll )
			{
				if ( (F = (STV_FIELDPTR) stv_findField(
						&STVFIELDS, what )) == (STV_FIELDPTR) NULL )
				{
					sprintf( STVERRTXT, "Unknown Field ID %d", what );

					STV_ERROR( "stv_setopt", stvStatusBadParam,
							STVERRTXT );

					oldvalue = -1;

					break;
				}	

				oldvalue = stv_time_of( F->time_base, F->time_units );

				stv_incr_time( &(F->time_base), &(F->time_units),
						tval, tunits );
			}

			else
			{
				F = STVFIELDS.next;

				oldvalue = -1;

				while ( F != NULL )
				{
					/* return the last field's "old value"... */
					oldvalue = stv_time_of( F->time_base,
							F->time_units );

					stv_incr_time( &(F->time_base), &(F->time_units),
							tval, tunits );

					F = F->next;
				}
			}

			break;
		}

		case stvOptFieldReady:
		{
			if ( what != stvOptAll )
			{
				if ( (F = (STV_FIELDPTR) stv_findField(
						&STVFIELDS, what )) == (STV_FIELDPTR) NULL )
				{
					sprintf( STVERRTXT, "Unknown Field ID %d", what );

					STV_ERROR( "stv_setopt", stvStatusBadParam,
							STVERRTXT );

					oldvalue = -1;

					break;
				}	

				oldvalue = F->data_ready;

				F->data_ready = ( value ) ? stvTrue : stvFalse;
			}

			else
			{
				F = STVFIELDS.next;

				oldvalue = -1;

				while ( F != NULL )
				{
					/* return the last field's "old value"... */
					oldvalue = F->data_ready;

					F->data_ready = ( value ) ? stvTrue : stvFalse;

					F = F->next;
				}
			}

			break;
		}

		default:
		{
			sprintf( STVERRTXT, "Unknown Option ID %d", option );

			STV_ERROR( "stv_setopt", stvStatusBadParam, STVERRTXT );

			oldvalue = -1;

			break;
		}
	}

	return( oldvalue );
}


int
stv_getopt( what, option )
int what;
int option;
{
	STV_FIELDPTR F;

	int value;

	switch ( option )
	{
		case stvOptFullUnpack:
		{
			value = STV_OPT_FULL_UNPACK;

			break;
		}

		case stvOptVerbose:
		{
			value = STV_OPT_VERBOSE;

			break;
		}

		case stvOptDumpAll:
		{
			value = STV_OPT_DUMP_ALL;

			break;
		}

		case stvOptDump:
		{
			value = STV_OPT_DUMP;

			break;
		}

		case stvOptAppLock:
		{
			value = STV_APP_LOCK;

			break;
		}

		case stvOptCpRecovery:
		{
			value = STV_OPT_CP_RECOVERY;

			break;
		}

		case stvOptFieldStorage:
		{
			if ( what == stvOptAll )
			{
				sprintf( STVERRTXT, "Invalid Arg \"%s\" to \"%s\"",
					"stvOptAll", "stvOptFieldStorage" );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}

			if ( (F = (STV_FIELDPTR) stv_findField( &STVFIELDS, what ))
					== (STV_FIELDPTR) NULL )
			{
				sprintf( STVERRTXT, "Unknown Field ID %d", what );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}	

			value = F->storage_order;

			if ( value != stvRowMajor && value != stvColumnMajor )
			{
				sprintf( STVERRTXT, "Unknown Storage Order %d",
					value );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );
			}
				
			break;
		}

		case stvOptFieldTimeScale:
		{
			if ( what == stvOptAll )
			{
				sprintf( STVERRTXT, "Invalid Arg \"%s\" to \"%s\"",
					"stvOptAll", "stvOptFieldTimeScale" );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}

			if ( (F = (STV_FIELDPTR) stv_findField( &STVFIELDS, what ))
					== (STV_FIELDPTR) NULL )
			{
				sprintf( STVERRTXT, "Unknown Field ID %d", what );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}	

			value = F->time_scale;

			if ( !stv_is_time_value( value ) )
			{
				sprintf( STVERRTXT, "Invalid Time Value %d", value );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );
			}
				
			break;
		}

		case stvOptSetFieldTime:
		{
			if ( what == stvOptAll )
			{
				sprintf( STVERRTXT, "Invalid Arg \"%s\" to \"%s\"",
					"stvOptAll", "stvOptSetFieldTime" );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}

			if ( (F = (STV_FIELDPTR) stv_findField( &STVFIELDS, what ))
					== (STV_FIELDPTR) NULL )
			{
				sprintf( STVERRTXT, "Unknown Field ID %d", what );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}	

			value = stv_time_of( F->time_base, F->time_units );

			break;
		}

		case stvOptIncrFieldTime:
		{
			if ( what == stvOptAll )
			{
				sprintf( STVERRTXT, "Invalid Arg \"%s\" to \"%s\"",
					"stvOptAll", "stvOptIncrFieldTime" );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}

			if ( (F = (STV_FIELDPTR) stv_findField( &STVFIELDS, what ))
					== (STV_FIELDPTR) NULL )
			{
				sprintf( STVERRTXT, "Unknown Field ID %d", what );

				STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

				value = -1;

				break;
			}	

			sprintf( STVERRTXT, "%s is not a readable option!",
				"stvOptIncrFieldTime" );

			STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

			value = -1;

			break;
		}

		case stvOptFieldReady:
		{
			if ( what != stvOptAll )
			{
				if ( (F = (STV_FIELDPTR) stv_findField(
						&STVFIELDS, what )) == (STV_FIELDPTR) NULL )
				{
					sprintf( STVERRTXT, "Unknown Field ID %d", what );

					STV_ERROR( "stv_getopt", stvStatusBadParam,
							STVERRTXT );

					value = -1;

					break;
				}	

				value = F->data_ready;
			}

			/* Check all fields for ready status */
			else
			{
				F = STVFIELDS.next;

				value = stvTrue;

				while ( F != NULL )
				{
					if ( !(F->data_ready) )
					{
						value = stvFalse;

						break;
					}

					F = F->next;
				}
			}

			break;
		}

		default:
		{
			sprintf( STVERRTXT, "Unknown Option ID %d", option );

			STV_ERROR( "stv_getopt", stvStatusBadParam, STVERRTXT );

			value = -1;

			break;
		}
	}

	return( value );
}


STV_DECOMPPTR
stv_create_decomp()
{
	STV_DECOMPPTR tmp;

	tmp = (STV_DECOMPPTR) MALLOC( sizeof( struct stv_decomp_struct ) );
	STV_MEMCHECK( tmp, "Decomposition Structure",
		return( (STV_DECOMPPTR) NULL ) );

    tmp->decompID = -1;
    tmp->dataDim = -1;

    tmp->axisType = (int *) NULL;

    tmp->axisInfo = (int *) NULL;

    tmp->axisLBounds = (int **) NULL;
    tmp->axisUBounds = (int **) NULL;

	tmp->nbounds = (int *) NULL;

    tmp->glb = (int *) NULL;
    tmp->gub = (int *) NULL;

    tmp->prank = -1;

    tmp->pshape = (int *) NULL;

    tmp->ntasks = -1;

    tmp->refcount = -1;

    tmp->prev = (STV_DECOMPPTR) NULL;
    tmp->next = (STV_DECOMPPTR) NULL;

	return( tmp );
}


void
stv_free_decomp( ptr )
STV_DECOMPPTR *ptr;
{
	STV_DECOMPPTR D;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
    D = *ptr;

    (D->refcount)--;

    if ( D->refcount > 0 )  /* only purge when refcount is 0 */ 
    {
       *ptr = (STV_DECOMPPTR) NULL;

       return;     
    }

    D->decompID = -1;

    if ( D->axisType != NULL )
	{
		FREE( D->axisType );

		D->axisType = (int *) NULL;
	}

    if ( D->axisInfo != NULL )
	{
		FREE( D->axisInfo );

    	D->axisInfo = (int *) NULL;
	}

    if ( D->axisLBounds != NULL )
	{
		for ( i=0 ; i < D->dataDim ; i++ )
		{
			if ( D->axisLBounds[i] != NULL )
			{
				FREE( D->axisLBounds[i] );

				D->axisLBounds[i] = (int *) NULL;
			}
		}

		FREE( D->axisLBounds );

    	D->axisLBounds = (int **) NULL;
	}

    if ( D->axisUBounds != NULL )
	{
		for ( i=0 ; i < D->dataDim ; i++ )
		{
			if ( D->axisUBounds[i] != NULL )
			{
				FREE( D->axisUBounds[i] );

				D->axisUBounds[i] = (int *) NULL;
			}
		}

		FREE( D->axisUBounds );

    	D->axisUBounds = (int **) NULL;
	}

    if ( D->nbounds != NULL )
	{
		FREE( D->nbounds );

    	D->nbounds = (int *) NULL;
	}

    if ( D->glb != NULL )
	{
		FREE( D->glb );

    	D->glb = (int *) NULL;
	}

    if ( D->gub != NULL )
	{
		FREE( D->gub );

    	D->gub = (int *) NULL;
	}

    D->dataDim = -1;
    D->prank = -1;

	if ( D->pshape != NULL )
	{
		FREE( D->pshape );

    	D->pshape = (int *) NULL;
	}

    D->ntasks = -1;

	D->refcount = -1;

    D->prev = (STV_DECOMPPTR) NULL;
    D->next = (STV_DECOMPPTR) NULL;

	FREE( D );

	*ptr = (STV_DECOMPPTR) NULL;
}


int
stv_match_decomp( D1, D2 )
STV_DECOMPPTR D1, D2;
{
	int i;

	if ( D1 == NULL && D2 == NULL )
		return( stvTrue );
	
	if ( D1 == NULL || D2 == NULL )
		return( stvFalse );

	/* Don't check decompID -- just too easy... */

	if ( D1->dataDim != D2->dataDim )
		return( stvFalse );

	for ( i=0 ; i < D1->dataDim ; i++ )
	{
		if ( D1->axisType[i] != D2->axisType[i] )
			return( stvFalse );

		if ( D1->axisType[i] != stvExplicit )
		{
			if ( D1->axisInfo[i] != D2->axisInfo[i] )
				return( stvFalse );
		}

		if ( D1->glb[i] != D2->glb[i] )
			return( stvFalse );

		if ( D1->gub[i] != D2->gub[i] )
			return( stvFalse );
	}

	if ( D1->prank != D2->prank )
		return( stvFalse );

	for ( i=0 ; i < D1->prank ; i++ )
	{
		if ( D1->pshape[i] != D2->pshape[i] )
			return( stvFalse );
	}

	if ( D1->ntasks != D2->ntasks )
		return( stvFalse );

	return( stvTrue );
}


STV_DECOMPPTR
stv_get_decomp( D, decomps, ndecomps )
STV_DECOMPPTR D;
STV_DECOMPPTR decomps;
int ndecomps;
{
	STV_DECOMPPTR tmp;

	tmp = decomps;

	while ( tmp != NULL )
	{
		if ( stv_match_decomp( D, tmp ) )
			return( tmp );

		tmp = tmp->next;
	}

	return( tmp );
}


STV_PARTICLEPTR
stv_create_particle()
{
	STV_PARTICLEPTR tmp;

	tmp = (STV_PARTICLEPTR) MALLOC(
		sizeof( struct stv_particle_struct ) );
	STV_MEMCHECK( tmp, "Particle Structure",
		return( (STV_PARTICLEPTR) NULL ) );

    tmp->particleID = -1;

	tmp->name = (char *) NULL;

	tmp->get_particle = (STV_GET_PARTICLE) NULL;

	tmp->make_particle = (STV_MAKE_PARTICLE) NULL;

    tmp->dataDim = -1;

    tmp->glb = (int *) NULL;
    tmp->gub = (int *) NULL;

    tmp->ntasks = -1;

	tmp->offset = -1;

    tmp->refcount = -1;

    tmp->prev = (STV_PARTICLEPTR) NULL;
    tmp->next = (STV_PARTICLEPTR) NULL;

	return( tmp );
}


void
stv_free_particle( ptr )
STV_PARTICLEPTR *ptr;
{
	STV_PARTICLEPTR PT;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
    PT = *ptr;

    (PT->refcount)--;

    if ( PT->refcount > 0 )  /* only purge when refcount is 0 */ 
    {
       *ptr = (STV_PARTICLEPTR) NULL;

       return;     
    }

    PT->particleID = -1;

	if ( PT->name != NULL )
	{
		FREE( PT->name );

		PT->name = (char *) NULL;
	}

	PT->get_particle = (STV_GET_PARTICLE) NULL;

	PT->make_particle = (STV_MAKE_PARTICLE) NULL;

    PT->dataDim = -1;

    if ( PT->glb != NULL )
	{
		FREE( PT->glb );

    	PT->glb = (int *) NULL;
	}

    if ( PT->gub != NULL )
	{
		FREE( PT->gub );

    	PT->gub = (int *) NULL;
	}

    PT->ntasks = -1;

	PT->offset = -1;

	PT->refcount = -1;

    PT->prev = (STV_PARTICLEPTR) NULL;
    PT->next = (STV_PARTICLEPTR) NULL;

	FREE( PT );

	*ptr = (STV_PARTICLEPTR) NULL;
}


int
stv_match_particle( PT1, PT2 )
STV_PARTICLEPTR PT1;
STV_PARTICLEPTR PT2;
{
	int i;

	if ( PT1 == NULL && PT2 == NULL )
		return( stvTrue );
	
	if ( PT1 == NULL || PT2 == NULL )
		return( stvFalse );

	/* Don't check particleID -- just too easy... */

	if ( PT1->name != NULL && PT2->name != NULL )
	{
		if ( strcmp( PT1->name, PT2->name ) )
			return( stvFalse );
	}

	else if ( PT1->name == NULL || PT2->name == NULL )
		return( stvFalse );

	if ( PT1->dataDim != PT2->dataDim )
		return( stvFalse );

	for ( i=0 ; i < PT1->dataDim ; i++ )
	{
		if ( PT1->glb[i] != PT2->glb[i] )
			return( stvFalse );

		if ( PT1->gub[i] != PT2->gub[i] )
			return( stvFalse );
	}

	if ( PT1->ntasks != PT2->ntasks )
		return( stvFalse );

	return( stvTrue );
}


STV_PARTICLEPTR
stv_get_particle_name( name, particles, nparticles )
char *name;
STV_PARTICLEPTR particles;
int nparticles;
{
	STV_PARTICLEPTR PT;

	PT = particles;

	while ( PT != NULL )
	{
		if ( !strcmp( PT->name, name ) )
			return( PT );

		PT = PT->next;
	}

	return( PT );
}


STV_PARTICLEPTR
stv_get_particle_name_vfg( VFG, name )
STV_VIEW_FIELD_GROUP VFG;
char *name;
{
	STV_VIEW_FIELD VF;

	int i;

	for ( i=0 ; i < VFG->nvfields ; i++ )
	{
		VF = VFG->view_field_list[i];

		if ( VF->field->particle != NULL
			&& !strcmp( VF->field->particle->name, name ) )
		{
			return( VF->field->particle );
		}
	}

	return( (STV_PARTICLEPTR) NULL );
}


STV_PARTICLEPTR
stv_get_particle( PT, particles, nparticles )
STV_PARTICLEPTR PT;
STV_PARTICLEPTR particles;
int nparticles;
{
	STV_PARTICLEPTR tmp;

	tmp = particles;

	while ( tmp != NULL )
	{
		if ( stv_match_particle( PT, tmp ) )
			return( tmp );

		tmp = tmp->next;
	}

	return( tmp );
}


int
stv_particle_in_region( R, coords, dim )
STV_REGION R;
int *coords;
int dim;
{
	int i;

	for ( i=0 ; i < dim ; i++ )
	{
		if ( STV_REGION_MIN( R, i ) > coords[i]
			|| STV_REGION_MAX( R, i ) < coords[i] )
		{
			return( stvFalse );
		}
	}

	return( stvTrue );
}


STV_FIELDPTR
stv_create_field()
{
	STV_FIELDPTR tmp;

	tmp = (STV_FIELDPTR) MALLOC( sizeof( struct stv_field_struct ) );
	STV_MEMCHECK( tmp, "Field Structure",
		return( (STV_FIELDPTR) NULL ) );

    tmp->fieldID = -1;

	tmp->name = (char *) NULL;

	tmp->decompID = -1;

    tmp->decomp = (STV_DECOMPPTR) NULL;

	tmp->particleID = -1;

    tmp->particle = (STV_PARTICLEPTR) NULL;

	tmp->get_pfield = (STV_GET_PFIELD) NULL;
	tmp->get_pfield_arg = (STV_VALUE) NULL;

	tmp->set_pfield = (STV_SET_PFIELD) NULL;
	tmp->set_pfield_arg = (STV_VALUE) NULL;

    tmp->var = (STV_VALUE) NULL;

    tmp->type = -1;

	tmp->nelems = -1;

	tmp->arrayOffset = (int *) NULL;

    tmp->arrayDecl = (int *) NULL;

    tmp->storage_order = -1;

    tmp->time_scale = -1;

	tmp->time_base = -1;

	tmp->time_units = -1;

	tmp->data_ready = -1;

    tmp->paddr = (int *) NULL;

	tmp->aflag = -1;

	tmp->offset = -1;

    tmp->refcount = -1;

	tmp->vlist = (STV_VLISTPTR) NULL;

    tmp->prev = (STV_FIELDPTR) NULL;
    tmp->next = (STV_FIELDPTR) NULL;

	return( tmp );
}


void
stv_free_field( ptr )
STV_FIELDPTR *ptr;
{
	STV_FIELDPTR F;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	F = *ptr;

    (F->refcount)--;

    if ( F->refcount > 0 )
    {
        *ptr = (STV_FIELDPTR) NULL;

        return;
    }

    F->fieldID = -1;

	if ( F->name != NULL )
	{
		FREE( F->name );

    	F->name = (char *) NULL;
	}

	F->decompID = -1;

    F->decomp = (STV_DECOMPPTR) NULL;

	F->particleID = -1;

    F->particle = (STV_PARTICLEPTR) NULL;

	F->get_pfield = (STV_GET_PFIELD) NULL;
	F->get_pfield_arg = (STV_VALUE) NULL;

	F->set_pfield = (STV_SET_PFIELD) NULL;
	F->set_pfield_arg = (STV_VALUE) NULL;

    F->var = (STV_VALUE) NULL;

    F->type = -1;

	F->nelems = -1;

	if ( F->arrayOffset != NULL )
	{
		FREE( F->arrayOffset );

    	F->arrayOffset = (int *) NULL;
	}

	if ( F->arrayDecl != NULL )
	{
		FREE( F->arrayDecl );

    	F->arrayDecl = (int *) NULL;
	}

    F->storage_order = -1;

    F->time_scale = -1;

	F->time_base = -1;

	F->time_units = -1;

	F->data_ready = -1;

    F->paddr = (int *) NULL;

	F->aflag = -1;

	F->offset = -1;

	F->refcount = -1;

    FREELIST( F->vlist, prev, next );

    F->prev = (STV_FIELDPTR) NULL;
    F->next = (STV_FIELDPTR) NULL;

    FREE( F );

    *ptr = (STV_FIELDPTR) NULL;
}


int
stv_match_field( F1, F2 )
STV_FIELDPTR F1, F2;
{
	int i;

	if ( F1 == NULL && F2 == NULL )
		return( stvTrue );
	
	if ( F1 == NULL || F2 == NULL )
		return( stvFalse );

	/* Only check "real" data, no IDs, ptrs, etc... */

	if ( F1->name != NULL && F2->name != NULL )
	{
		if ( strcmp( F1->name, F2->name ) )
			return( stvFalse );
	}

	else if ( F1->name == NULL || F2->name == NULL )
		return( stvFalse );

	if ( !stv_match_decomp( F1->decomp, F2->decomp ) )
		return( stvFalse );

	if ( !stv_match_particle( F1->particle, F2->particle ) )
		return( stvFalse );

	if ( F1->type != F2->type )
		return( stvFalse );

	if ( F1->nelems != F2->nelems )
		return( stvFalse );

	if ( F1->storage_order != F2->storage_order )
		return( stvFalse );

	if ( F1->time_scale != F2->time_scale )
		return( stvFalse );

	/* Don't compare time_base or time_units - dynamic values... */

	return( stvTrue );
}


STV_FIELDPTR
stv_get_field_name( name, fields, nfields )
char *name;
STV_FIELDPTR fields;
int nfields;
{
	STV_FIELDPTR F;

	F = fields;

	while ( F != NULL )
	{
		if ( !strcmp( F->name, name ) )
			return( F );

		F = F->next;
	}

	return( F );
}


STV_FIELDPTR
stv_get_field_name_array( name, fields, nfields )
char *name;
STV_FIELDPTR *fields;
int nfields;
{
	STV_FIELDPTR F;

	int i;

	for ( i=0 ; i < nfields ; i++ )
	{
		F = fields[i];

		if ( !strcmp( F->name, name ) )
			return( F );
	}

	return( (STV_FIELDPTR) NULL );
}


int
stv_field_region_overlap( F, R, C )
STV_FIELDPTR F;
STV_REGION R;
STV_CELL C;
{
	STV_PACK_DIM PD;

	/* No Static Regions for Particle Fields... */

	if ( F->particle != NULL )
		return( stvTrue );
	
	/* Decomposed Field */

	else if ( F->decomp != NULL )
	{
		/* Check Packing Space */

		PD = stv_get_SubRegion_indices( F, R, C );

		if ( PD != NULL )
		{
			stv_free_pack_dim_list( &PD );

			return( stvTrue );
		}

		else
			return( stvFalse );
	}

	/* Unknown Field Type */

	else
	{
		sprintf( STVERRTXT, "Field \"%s\" Has No Decomp or Particle!",
			F->name );
		
		STV_ERROR( "stv_field_region_overlap", stvStatusBadParam,
			STVERRTXT );

		return( stvFalse );
	}
}


STV_FIELD_GROUP_PTR
stv_create_field_group()
{
	STV_FIELD_GROUP_PTR tmp;

	tmp = (STV_FIELD_GROUP_PTR) MALLOC(
		sizeof( struct stv_field_group_struct ) );
	STV_MEMCHECK( tmp, "Field Group Structure",
		return( (STV_FIELD_GROUP_PTR) NULL ) );

	tmp->fgid = -1;

	tmp->field_list = (STV_FIELDPTR *) NULL;
	tmp->field_types = (int *) NULL;
	tmp->field_storage_orders = (int *) NULL;
	tmp->field_ids = (int *) NULL;
	tmp->nfields = -1;

    tmp->vRegionId = -1;
    tmp->vState = -1;

	stv_init_region( tmp->visRegion );

	stv_init_region( tmp->visSubregion );

	stv_init_cell( tmp->visCell );

	tmp->OKtoSend = -1;
	tmp->NeedSend = -1;
    tmp->toffset = -1;
	tmp->eff_freq = -1;
	tmp->freq = stvNotTime;
	tmp->newEffFreq = -1;
	tmp->newFreq = stvNotTime;
	tmp->nfTimeStamp = -1;
	tmp->no_synch = -1;
	tmp->is_static = -1;

    tmp->prev = (STV_FIELD_GROUP_PTR) NULL;
    tmp->next = (STV_FIELD_GROUP_PTR) NULL;

	return( tmp );
}


void
stv_free_field_group( ptr )
STV_FIELD_GROUP_PTR *ptr;
{
	STV_FIELD_GROUP_PTR FG;

	int i;

    if ( ptr == NULL || *ptr == NULL )
		return;

	FG = *ptr;

	FG->fgid = -1;

	if ( FG->field_list != NULL )
	{
		for ( i=0 ; i < FG->nfields ; i++ )
			FG->field_list[i] = (STV_FIELDPTR) NULL;

		FREE( FG->field_list );

		FG->field_list = (STV_FIELDPTR *) NULL;
	}

	if ( FG->field_types != NULL )
	{
		for ( i=0 ; i < FG->nfields ; i++ )
			FG->field_types[i] = -1;

		FREE( FG->field_types );

		FG->field_types = (int *) NULL;
	}

	if ( FG->field_storage_orders != NULL )
	{
		for ( i=0 ; i < FG->nfields ; i++ )
			FG->field_storage_orders[i] = -1;

		FREE( FG->field_storage_orders );

		FG->field_storage_orders = (int *) NULL;
	}

	if ( FG->field_ids != NULL )
	{
		for ( i=0 ; i < FG->nfields ; i++ )
			FG->field_ids[i] = -1;

		FREE( FG->field_ids );

		FG->field_ids = (int *) NULL;
	}

	FG->nfields = -1;

    FG->vRegionId = -1;
    FG->vState = -1;

	stv_init_region( FG->visRegion );

	stv_init_region( FG->visSubregion );

	stv_init_cell( FG->visCell );

	FG->OKtoSend = -1;
	FG->NeedSend = -1;
    FG->toffset = -1;
	FG->eff_freq = -1;
	FG->freq = stvNotTime;
	FG->newEffFreq = -1;
	FG->newFreq = stvNotTime;
	FG->nfTimeStamp = -1;
	FG->no_synch = -1;
	FG->is_static = -1;

    FG->prev = (STV_FIELD_GROUP_PTR) NULL;
    FG->next = (STV_FIELD_GROUP_PTR) NULL;

	FREE( FG );

	*ptr = (STV_FIELD_GROUP_PTR) NULL;
}


void
stv_free_field_group_list( ptr )
STV_FIELD_GROUP_PTR *ptr;
{
	STV_FIELD_GROUP_PTR FGnext;
	STV_FIELD_GROUP_PTR FG;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	FG = *ptr;

	while ( FG != NULL )
	{
		FGnext = FG->next;

		stv_free_field_group( &FG );

		FG = FGnext;
	}

	*ptr = (STV_FIELD_GROUP_PTR) NULL;
}


STV_FIELD_MAP_PTR
stv_create_field_map()
{
    STV_FIELD_MAP_PTR tmp;

    tmp = (STV_FIELD_MAP_PTR) MALLOC(
		sizeof( struct stv_field_map_struct ) );
    STV_MEMCHECK( tmp, "Field Mapping Structure",
		return( (STV_FIELD_MAP_PTR) NULL ) );

    tmp->remoteID = -1;

    tmp->local = (STV_FIELDPTR) NULL;

    tmp->prev = (STV_FIELD_MAP_PTR) NULL;
    tmp->next = (STV_FIELD_MAP_PTR) NULL;

    return( tmp );
}


void
stv_free_field_map( ptr )
STV_FIELD_MAP_PTR *ptr;
{
    STV_FIELD_MAP_PTR FM;

    if ( ptr == NULL || *ptr == NULL )
       return;
	
    FM = *ptr;

    FM->remoteID = -1;

	if ( FM->local != NULL )
		stv_free_field( &(FM->local) );

    FM->prev = (STV_FIELD_MAP_PTR) NULL;
    FM->next = (STV_FIELD_MAP_PTR) NULL;

    FREE( FM );

    *ptr = (STV_FIELD_MAP_PTR) NULL;
}


void
stv_free_field_map_list( ptr )
STV_FIELD_MAP_PTR *ptr;
{
	STV_FIELD_MAP_PTR FMnext;
	STV_FIELD_MAP_PTR FM;

	if ( ptr == NULL || *ptr == NULL )
		return;

	FM = *ptr;

	while ( FM != NULL )
	{
		FMnext = FM->next;

		stv_free_field_map( &FM );

		FM = FMnext;
	}

	*ptr = (STV_FIELD_MAP_PTR) NULL;
}


STV_FIELD_MAP_PTR
stv_get_field_map( remoteID, field_maps )
int remoteID;
STV_FIELD_MAP_PTR field_maps;
{
	STV_FIELD_MAP_PTR FM;

	FM = field_maps;

	while ( FM != NULL )
	{
		if ( FM->remoteID == remoteID )
			return( FM );

		FM = FM->next;
	}

	return( FM );
}


STV_PARAMPTR
stv_create_param()
{
	STV_PARAMPTR tmp;

	tmp = (STV_PARAMPTR) MALLOC( sizeof( struct stv_param_struct ) );
	STV_MEMCHECK( tmp, "Parameter Structure",
		return( (STV_PARAMPTR) NULL ) );

    tmp->paramID = -1;

	tmp->name = (char *) NULL;

    tmp->pnames = (char **) NULL;

    tmp->vars = (STV_VALUE *) NULL;

    tmp->types = (int *) NULL;

	tmp->num = -1;

	tmp->aflag = -1;

    tmp->refcount = -1;

	tmp -> updated = 0;

    tmp->prev = (STV_PARAMPTR) NULL;
    tmp->next = (STV_PARAMPTR) NULL;

	return( tmp );
}


void
stv_free_param( ptr )
STV_PARAMPTR *ptr;
{
	STV_PARAMPTR P;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;

	P = *ptr;

	(P->refcount)--;

	if ( P->refcount > 0 )
	{
		*ptr = (STV_PARAMPTR) NULL;

		return;
	}

	P->paramID = -1;

	if ( P->name != NULL )
	{
		FREE( P->name );

		P->name = (char *) NULL;
	}

	for ( i=0 ; i < P->num ; i++ )
	{
		if ( P->pnames != NULL && P->pnames[i] != NULL )
		{
			FREE( P->pnames[i] );

			P->pnames[i] = (char *) NULL;
		}

		if ( P->vars != NULL )
			P->vars[i] = (STV_VALUE) NULL;

		if ( P->types != NULL )
			P->types[i] = -1;
	}

	if ( P->pnames != NULL )
	{
		FREE( P->pnames );

		P->pnames = (char **) NULL;
	}

	if ( P->vars != NULL )
	{
		FREE( P->vars );

		P->vars = (STV_VALUE *) NULL;
	}

	if ( P->types != NULL )
	{
		FREE( P->types );

		P->types = (int *) NULL;
	}

	P->num = -1;

	P->aflag = -1;

	P->refcount = -1;

	P->prev = (STV_PARAMPTR) NULL;
	P->next = (STV_PARAMPTR) NULL;

	FREE( P );

	*ptr = (STV_PARAMPTR) NULL;
}


int
stv_match_param( P1, P2 )
STV_PARAMPTR P1, P2;
{
	int i;

	if ( P1 == NULL && P2 == NULL )
		return( stvTrue );
	
	if ( P1 == NULL || P2 == NULL )
		return( stvFalse );

	/* Only check "real" data, no IDs, ptrs, etc... */

	if ( P1->name != NULL && P2->name != NULL )
	{
		if ( strcmp( P1->name, P2->name ) )
			return( stvFalse );
	}

	else if ( P1->name == NULL || P2->name == NULL )
		return( stvFalse );

	if ( P1->num != P2->num )
		return( stvFalse );
	
	if ( P1->num > 0 )
	{
		if ( P1->pnames != NULL && P2->pnames != NULL
			&& P1->types != NULL && P2->types != NULL )
		{
			for ( i=0 ; i < P1->num ; i++ )
			{
				if ( strcmp( P1->pnames[i], P2->pnames[i] ) )
					return( stvFalse );

				if ( P1->types[i] != P2->types[i] )
					return( stvFalse );
			}
		}

		else
			return( stvFalse );
	}

	return( stvTrue );
}


STV_PARAMPTR
stv_get_param_name( name, params, nparams )
char *name;
STV_PARAMPTR params;
int nparams;
{
	STV_PARAMPTR P;

	P = params;

	while ( P != NULL )
	{
		if ( !strcmp( P->name, name ) )
			return( P );

		P = P->next;
	}

	return( P );
}


STV_VALUE_LIST
stv_create_value_list()
{
    STV_VALUE_LIST tmp;

    tmp = (STV_VALUE_LIST) MALLOC(
		sizeof( struct stv_value_list_struct ) );
    STV_MEMCHECK( tmp, "Value List Structure",
		return( (STV_VALUE_LIST) NULL ) );

	tmp->values = (STV_VALUE *) NULL;

	tmp->num = -1;

    tmp->tvalid = -1;

	tmp->update_fgid = -1;

    tmp->next = (STV_VALUE_LIST) NULL;

    return( tmp );
}


void
stv_free_value_list( ptr )
STV_VALUE_LIST *ptr;
{
    STV_VALUE_LIST VL;

	int i;

    if ( ptr == NULL || *ptr == NULL )
       return;
	
    VL = *ptr;

	if ( VL->values != NULL )
	{
		for ( i=0 ; i < VL->num ; i++ )
		{
			if ( VL->values[i] != NULL )
			{
				FREE( VL->values[i] );

				VL->values[i] = (STV_VALUE) NULL;
			}
		}

		FREE( VL->values );

		VL->values = (STV_VALUE *) NULL;
	}

	VL->num = -1;

    VL->tvalid = -1;

	VL->update_fgid = -1;

    VL->next = (STV_VALUE_LIST) NULL;

    FREE( VL );

    *ptr = (STV_VALUE_LIST) NULL;
}


STV_PARAM_MAP_PTR
stv_create_param_map()
{
    STV_PARAM_MAP_PTR tmp;

    tmp = (STV_PARAM_MAP_PTR) MALLOC(
		sizeof( struct stv_param_map_struct ) );
    STV_MEMCHECK( tmp, "Parameter Mapping Structure",
		return( (STV_PARAM_MAP_PTR) NULL ) );

    tmp->remoteID = -1;

    tmp->local = (STV_PARAMPTR) NULL;

	tmp->newVals = (STV_VALUE_LIST) NULL;

    tmp->prev = (STV_PARAM_MAP_PTR) NULL;
    tmp->next = (STV_PARAM_MAP_PTR) NULL;

    return( tmp );
}


void
stv_free_param_map( ptr )
STV_PARAM_MAP_PTR *ptr;
{
    STV_PARAM_MAP_PTR PM;

	int i;

    if ( ptr == NULL || *ptr == NULL )
       return;
	
    PM = *ptr;

    PM->remoteID = -1;

	if ( PM->local != NULL )
		stv_free_param( &(PM->local) );

	if ( PM->newVals != NULL )
		stv_free_value_list( &(PM->newVals) );

    PM->prev = (STV_PARAM_MAP_PTR) NULL;
    PM->next = (STV_PARAM_MAP_PTR) NULL;

    FREE( PM );

    *ptr = (STV_PARAM_MAP_PTR) NULL;
}


void
stv_free_param_map_list( ptr )
STV_PARAM_MAP_PTR *ptr;
{
	STV_PARAM_MAP_PTR PMnext;
	STV_PARAM_MAP_PTR PM;

	if ( ptr == NULL || *ptr == NULL )
		return;

	PM = *ptr;

	while ( PM != NULL )
	{
		PMnext = PM->next;

		stv_free_param_map( &PM );

		PM = PMnext;
	}

	*ptr = (STV_PARAM_MAP_PTR) NULL;
}


STV_PARAM_MAP_PTR
stv_get_param_map( remoteID, param_maps )
int remoteID;
STV_PARAM_MAP_PTR param_maps;
{
	STV_PARAM_MAP_PTR PM;

	PM = param_maps;

	while ( PM != NULL )
	{
		if ( PM->remoteID == remoteID )
			return( PM );

		PM = PM->next;
	}

	return( PM );
}


STV_FE_VIEWERPTR
stv_create_fe_viewer()
{
	STV_FE_VIEWERPTR tmp;

	tmp = (STV_FE_VIEWERPTR) MALLOC(
		sizeof( struct stv_fe_viewer_struct ) );
	STV_MEMCHECK( tmp, "Viewer Structure",
		return( (STV_FE_VIEWERPTR) NULL ) );

    tmp->vtid = -1;

	tmp->vid = -1;

    tmp->nfgs = -1;
	tmp->fg_list = (STV_FIELD_GROUP_PTR) NULL;

	tmp->field_map = (STV_FIELD_MAP_PTR) NULL;

    tmp->param_map = (STV_PARAM_MAP_PTR) NULL;

    tmp->nNewParams = -1;

    tmp->refcount = -1;

    tmp->prev = (STV_FE_VIEWERPTR) NULL;
    tmp->next = (STV_FE_VIEWERPTR) NULL;

	return( tmp );
}


void
stv_free_fe_viewer( ptr )
STV_FE_VIEWERPTR *ptr;
{
	STV_FE_VIEWERPTR V;

	int i;

    if ( ptr == NULL || *ptr == NULL )
		return;

	V = *ptr;

	(V->refcount)--;

	if ( V->refcount > 0 )
	{
		*ptr = (STV_FE_VIEWERPTR) NULL;

		return;
	}

    V->vtid = -1;

	V->vid = -1;

	if ( V->fg_list != NULL )
		stv_free_field_group_list( &(V->fg_list) );

	if ( V->field_map != NULL )
    	stv_free_field_map_list( &(V->field_map) );

	if ( V->param_map != NULL )
    	stv_free_param_map_list( &(V->param_map) );

    V->nNewParams = -1;

    V->refcount = -1;

    V->prev = (STV_FE_VIEWERPTR) NULL;
    V->next = (STV_FE_VIEWERPTR) NULL;

	FREE( V );

	*ptr = (STV_FE_VIEWERPTR) NULL;
}


STV_BUFPTR
stv_create_mbuffer()
{
	STV_BUFPTR tmp;

	tmp = (STV_BUFPTR) MALLOC( sizeof( struct stv_mbuffer_struct ) );
	STV_MEMCHECK( tmp, "Message Buffer Structure",
		return( (STV_BUFPTR) NULL ) );

    tmp->bufid = -1;

	tmp->fgid = -1;

	tmp->op = -1;

    tmp->src = -1;
    tmp->tag = -1;

	tmp->ctxt = (STV_CONTEXT) NULL;

    tmp->prev = (STV_BUFPTR) NULL;
    tmp->next = (STV_BUFPTR) NULL;

	return( tmp );
}


void
stv_free_mbuffer( ptr )
STV_BUFPTR *ptr;
{
	STV_BUFPTR B;

	if ( ptr == NULL || *ptr == NULL )
		return;

	B = *ptr;

    B->bufid = -1;

	B->fgid = -1;

	B->op = -1;

    B->src = -1;
    B->tag = -1;

	B->ctxt = (STV_CONTEXT) NULL;

    B->prev = (STV_BUFPTR) NULL;
    B->next = (STV_BUFPTR) NULL;

	FREE( B );

	*ptr = (STV_BUFPTR) NULL;
}


STV_HANDLER
stv_create_handler()
{
	STV_HANDLER tmp;

	tmp = (STV_HANDLER) MALLOC( sizeof( struct stv_handler_struct ) );
	STV_MEMCHECK( tmp, "Message Handler Structure",
		return( (STV_HANDLER) NULL ) );

	tmp->handler = (STV_HANDLER_PROC) NULL;

	tmp->op = -1;

    tmp->src = -1;
    tmp->tag = -1;

	tmp->ctxt = (STV_CONTEXT) NULL;

    tmp->next = (STV_HANDLER) NULL;

	return( tmp );
}


void
stv_free_handler( ptr )
STV_HANDLER *ptr;
{
	STV_HANDLER H;

	if ( ptr == NULL || *ptr == NULL )
		return;

	H = *ptr;

    H->handler = (STV_HANDLER_PROC) NULL;

	H->op = -1;

    H->src = -1;
    H->tag = -1;

	H->ctxt = (STV_CONTEXT) NULL;

    H->next = (STV_HANDLER) NULL;

	FREE( H );

	*ptr = (STV_HANDLER) NULL;
}


int
stv_add_handler( handlers, nhandlers, hfunc, op, src, tag, ctxt )
STV_HANDLER *handlers;
int *nhandlers;
STV_HANDLER_PROC hfunc;
int op;
int src;
int tag;
STV_CONTEXT ctxt;
{
	STV_HANDLER Hlast;
	STV_HANDLER H;

	char tmp[255];

	/* Check for Conflicting Message Handlers */

	H = *handlers;

	Hlast = (STV_HANDLER) NULL;

	while ( H != NULL )
	{
		if ( ( H->op == -1 || op == -1 || H->op == op )
				&& ( H->src == -1 || src == -1 || H->src == src )
				&& ( H->tag == -1 || tag == -1 || H->tag == tag )
				&& STV_CONTEXT_MATCH( H->ctxt, ctxt ) )
		{
			STV_CONTEXT_SPRINTF( tmp, H->ctxt );

			sprintf( STVERRTXT, "%s (op=%d,src=%d,tag=%d,ctxt=%s)",
				"Conflicting Handler Exists!",
				H->op, H->src, H->tag, tmp );
			STV_ERROR( "stv_add_handler",
					stvStatusBadParam, STVERRTXT );

			return( stvStatusBadParam );
		}

		Hlast = H;

		H = H->next;
	}

	/* Else Add The Bastard :-) */

	H = stv_create_handler();

	STV_NO_MEM_return( H, "stv_add_handler" );

	H->handler = hfunc;

	H->op = op;

	H->src = src;
	H->tag = tag;

	H->ctxt = ctxt;

	if ( Hlast != NULL )
		Hlast->next = H;

	else
		*handlers = H;
	
	(*nhandlers)++;

	return( stvStatusOk );
}


int
stv_delete_handler( handlers, nhandlers, hfunc, op, src, tag, ctxt )
STV_HANDLER *handlers;
int *nhandlers;
STV_HANDLER_PROC hfunc;
int op;
int src;
int tag;
STV_CONTEXT ctxt;
{
	STV_HANDLER Hlast;
	STV_HANDLER H;

	/* Find Message Handler */

	H = *handlers;

	Hlast = (STV_HANDLER) NULL;

	while ( H != NULL )
	{
		if ( H->handler == hfunc && H->op == op && H->src == src
				&& H->tag == tag && STV_CONTEXT_MATCH( H->ctxt, ctxt ) )
		{
			if ( Hlast != NULL )
				Hlast->next = H->next;
			
			else
				*handlers = H->next;
			
			(*nhandlers)--;

			stv_free_handler( &H );

			return( stvStatusOk );
		}

		Hlast = H;

		H = H->next;
	}

	STV_ERROR( "stv_delete_handler",
			stvStatusBadParam, "Matching Handler Not Found" );

	return( stvStatusBadParam );
}


int
stv_vector_handler( handlers, nhandlers, msg )
STV_HANDLER handlers;
int nhandlers;
STV_BUFPTR msg;
{
	STV_HANDLER H;

	int handled = stvFalse;
	int cc;

	H = handlers;

	cc = stvStatusOk;

	while ( H != NULL && !handled )
	{
		if ( ( H->op == -1 || H->op == msg->op )
				&& ( H->src == -1 || H->src == msg->src )
				&& ( H->tag == -1 || H->tag == msg->tag )
				&& STV_CONTEXT_MATCH( H->ctxt, msg->ctxt ) )
		{
			cc = (H->handler)( msg );

			handled = stvTrue;
		}

		H = H->next;
	}

	if ( !handled )
	{
		sprintf( STVERRTXT,
				"No Handler Found for Message Type \"%s\" (%d)",
				stv_msg_op_str( msg->op ), msg->op );
		STV_ERROR( "stv_vector_handler", stvStatusBadParam, STVERRTXT );

		cc = stvStatusBadParam;
	}

	return( cc );
}


STV_MSG_PASS_THRU
stv_create_msg_pass_thru()
{
	STV_MSG_PASS_THRU tmp;

	tmp = (STV_MSG_PASS_THRU) MALLOC(
			sizeof( struct stv_msg_pass_thru_struct ) );
	STV_MEMCHECK( tmp, "Message Pass Thru Structure",
			return( (STV_MSG_PASS_THRU) NULL ) );

	tmp->op = -1;

    tmp->next = (STV_MSG_PASS_THRU) NULL;

	return( tmp );
}


void
stv_free_msg_pass_thru( ptr )
STV_MSG_PASS_THRU *ptr;
{
	STV_MSG_PASS_THRU PT;

	if ( ptr == NULL || *ptr == NULL )
		return;

	PT = *ptr;

	PT->op = -1;

    PT->next = (STV_MSG_PASS_THRU) NULL;

	FREE( PT );

	*ptr = (STV_MSG_PASS_THRU) NULL;
}


int
stv_add_msg_pass_thru( op )
int op;
{
	STV_MSG_PASS_THRU PT;

	PT = stv_create_msg_pass_thru();
	
	STV_NO_MEM_return( PT, "stv_add_msg_pass_thru" );

	PT->op = op;

	PT->next = STVMSGPASSTHRUS;

	STVMSGPASSTHRUS = PT;

	return( stvStatusOk );
}


int
stv_delete_msg_pass_thru( op )
int op;
{
	STV_MSG_PASS_THRU PTlast;
	STV_MSG_PASS_THRU PT;

	PT = STVMSGPASSTHRUS;

	PTlast = (STV_MSG_PASS_THRU) NULL;

	while ( PT != NULL )
	{
		if ( PT->op == op )
		{
			if ( PTlast != NULL )
				PTlast->next = PT->next;
			
			else
				STVMSGPASSTHRUS = PT->next;

			stv_free_msg_pass_thru( &PT );

			return( stvStatusOk );
		}

		PTlast = PT;

		PT = PT->next;
	}

	return( stvStatusBadParam );
}


STV_VIEW_FIELD
stv_create_view_field()
{
	STV_VIEW_FIELD tmp;

	tmp = (STV_VIEW_FIELD) MALLOC(
		sizeof( struct stv_view_field_struct ) );
	STV_MEMCHECK( tmp, "View Field Structure",
		return( (STV_VIEW_FIELD) NULL ) );

	tmp->field = (STV_FIELDPTR) NULL;

	tmp->view_type = -1;

	tmp->view_storage_order = -1;

	tmp->selected = -1;

	tmp->tids = (int *) NULL;

	tmp->ntids = -1;

	tmp->next = (STV_VIEW_FIELD) NULL;

	return( tmp );
}


void
stv_free_view_field( ptr )
STV_VIEW_FIELD *ptr;
{
	STV_VIEW_FIELD VF;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VF = *ptr;

	VF->field = (STV_FIELDPTR) NULL;

	VF->view_type = -1;

	VF->view_storage_order = -1;

	VF->selected = -1;

	if ( VF->tids != NULL )
	{
		FREE( VF->tids );

		VF->tids = (int *) NULL;
	}

	VF->ntids = -1;

	VF->next = (STV_VIEW_FIELD) NULL;

	FREE( VF );

	*ptr = (STV_VIEW_FIELD) NULL;
}


STV_VIEW_FIELD
stv_get_view_field_name( name, vfields, nvfields )
char *name;
STV_VIEW_FIELD vfields;
int nvfields;
{
	STV_VIEW_FIELD VF;

	VF = vfields;

	while ( VF != NULL )
	{
		if ( VF->field != NULL && !strcmp( VF->field->name, name ) )
			return( VF );

		VF = VF->next;
	}

	return( VF );
}


void
stv_clear_view_field_select( V )
STV_VIEWER V;
{
	STV_VIEW_FIELD VF;

	VF = V->vfields;

	while ( VF != NULL )
	{
		VF->selected = stvFalse;

		VF->view_type = -1;

		VF->view_storage_order = -1;

		VF = VF->next;
	}
}


STV_VIEW_PARTICLE_DATA
stv_create_view_particle_data()
{
	STV_VIEW_PARTICLE_DATA tmp;

	tmp = (STV_VIEW_PARTICLE_DATA) MALLOC(
		sizeof( struct stv_view_particle_data_struct ) );
	STV_MEMCHECK( tmp, "View Particle Data Structure",
		return( (STV_VIEW_PARTICLE_DATA) NULL ) );

	tmp->coords = (int *) NULL;

	tmp->field_values = (STV_VALUE *) NULL;

	tmp->field_nelems = (int *) NULL;

	tmp->next = (STV_VIEW_PARTICLE_DATA) NULL;

	return( tmp );
}


void
stv_free_view_particle_data( ptr )
STV_VIEW_PARTICLE_DATA *ptr;
{
	STV_VIEW_PARTICLE_DATA VPD;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VPD = *ptr;

	if ( VPD->coords != NULL )
	{
		FREE( VPD->coords );

		VPD->coords = (int *) NULL;
	}

	if ( VPD->field_values != NULL )
	{
		FREE( VPD->field_values );

		VPD->field_values = (STV_VALUE *) NULL;
	}

	if ( VPD->field_nelems != NULL )
	{
		FREE( VPD->field_nelems );

		VPD->field_nelems = (int *) NULL;
	}

	VPD->next = (STV_VIEW_PARTICLE_DATA) NULL;

	FREE( VPD );

	*ptr = (STV_VIEW_PARTICLE_DATA) NULL;
}


void
stv_free_view_particle_data_list( ptr )
STV_VIEW_PARTICLE_DATA *ptr;
{
	STV_VIEW_PARTICLE_DATA VPDnext;
	STV_VIEW_PARTICLE_DATA VPD;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VPD = *ptr;

	while ( VPD != NULL )
	{
		VPDnext = VPD->next;

		stv_free_view_particle_data( &VPD );

		VPD = VPDnext;
	}

	*ptr = (STV_VIEW_PARTICLE_DATA) NULL;
}


STV_VIEW_PARTICLE
stv_create_view_particle()
{
	STV_VIEW_PARTICLE tmp;

	tmp = (STV_VIEW_PARTICLE) MALLOC(
		sizeof( struct stv_view_particle_struct ) );
	STV_MEMCHECK( tmp, "View Particle Structure",
		return( (STV_VIEW_PARTICLE) NULL ) );

	tmp->particle = (STV_PARTICLEPTR) NULL;

	tmp->view_field_list = (STV_VIEW_FIELD *) NULL;

	tmp->view_types = (int *) NULL;

	tmp->nvfields = -1;

	tmp->pdata = (STV_VIEW_PARTICLE_DATA) NULL;

	tmp->npdata = -1;

	tmp->next = (STV_VIEW_PARTICLE) NULL;

	return( tmp );
}


void
stv_free_view_particle( ptr )
STV_VIEW_PARTICLE *ptr;
{
	STV_VIEW_PARTICLE VPT;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VPT = *ptr;

	VPT->particle = (STV_PARTICLEPTR) NULL;

	if ( VPT->view_field_list != NULL )
	{
		for ( i=0 ; i < VPT->nvfields ; i++ )
			VPT->view_field_list[i] = (STV_VIEW_FIELD) NULL;

		FREE( VPT->view_field_list );

		VPT->view_field_list = (STV_VIEW_FIELD *) NULL;
	}

	if ( VPT->view_types != NULL )
	{
		for ( i=0 ; i < VPT->nvfields ; i++ )
			VPT->view_types[i] = -1;

		FREE( VPT->view_types );

		VPT->view_types = (int *) NULL;
	}

	VPT->nvfields = -1;

	if ( VPT->pdata != NULL )
		stv_free_view_particle_data_list( &(VPT->pdata) );

	VPT->npdata = -1;

	VPT->next = (STV_VIEW_PARTICLE) NULL;

	FREE( VPT );

	*ptr = (STV_VIEW_PARTICLE) NULL;
}


void
stv_free_view_particle_list( ptr )
STV_VIEW_PARTICLE *ptr;
{
	STV_VIEW_PARTICLE VPTnext;
	STV_VIEW_PARTICLE VPT;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VPT = *ptr;

	while ( VPT != NULL )
	{
		VPTnext = VPT->next;

		stv_free_view_particle( &VPT );

		VPT = VPTnext;
	}

	*ptr = (STV_VIEW_PARTICLE) NULL;
}


void
stv_free_view_particle_fields( VPT )
STV_VIEW_PARTICLE VPT;
{
	STV_VIEW_PARTICLE_DATA VPD;

	int i;

	if ( STV_OPT_VERBOSE & STV_DEBUG_MEM )
	{
		printf( "Freeing Particle Data for Particle \"%s\"\n",
			VPT->particle->name );
		fflush( stdout );
	}

	VPD = VPT->pdata;

	while ( VPD != NULL )
	{
		for ( i=0 ; i < VPT->particle->dataDim ; i++ )
			VPD->coords[i] = -1;

		for ( i=0 ; i < VPT->nvfields ; i++ )
		{
			FREE( VPD->field_values[i] );

			VPD->field_values[i] = (STV_VALUE) NULL;
		}

		VPD = VPD->next;
	}
}


STV_VIEW_PARTICLE
stv_get_view_particle_name( VPT, name )
STV_VIEW_PARTICLE VPT;
char *name;
{
	STV_VIEW_PARTICLE VPTptr;

	VPTptr = VPT;

	while ( VPTptr != NULL )
	{
		if ( !strcmp( name, VPTptr->particle->name ) )
			return( VPTptr );

		VPTptr = VPTptr->next;
	}

	return( VPTptr );

}


STV_VIEW_FIELD_GROUP
stv_create_view_field_group()
{
	STV_VIEW_FIELD_GROUP tmp;

	tmp = (STV_VIEW_FIELD_GROUP) MALLOC(
		sizeof( struct stv_view_field_group_struct ) );
	STV_MEMCHECK( tmp, "View Field Group Structure",
		return( (STV_VIEW_FIELD_GROUP) NULL ) );

	tmp->fgid = -1;

	tmp->view_field_list = (STV_VIEW_FIELD *) NULL;

	tmp->view_types = (int *) NULL;

	tmp->view_storage_orders = (int *) NULL;

	tmp->field_values = (STV_VALUE *) NULL;

	tmp->field_arrayDecls = (int **) NULL;

	tmp->nvfields = -1;

	tmp->particle_list = (STV_VIEW_PARTICLE) NULL;

	tmp->tid_list = (int *) NULL;

	tmp->ntids = -1;

	stv_init_region( tmp->visregion );

	stv_init_cell( tmp->viscell );

	tmp->region_time = -1;;
	tmp->time_offset = -1;;
	tmp->region_cnt = -1;
	tmp->region_id = -1;
	tmp->reply_cnt = -1;
	tmp->frame_ok = -1;
	tmp->ack_cnt = -1;
	tmp->timemax = -1;
	tmp->locked = -1;
	tmp->eff_freq = -1;
	tmp->freq = stvNotTime;
	tmp->flag = -1;
	tmp->bail = -1;

	tmp->next = (STV_VIEW_FIELD_GROUP) NULL;

	return( tmp );
}


void
stv_free_view_field_group( ptr )
STV_VIEW_FIELD_GROUP *ptr;
{
	STV_VIEW_FIELD_GROUP VFG;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VFG = *ptr;

	VFG->fgid = -1;

	if ( VFG->view_field_list != NULL )
	{
		for ( i=0 ; i < VFG->nvfields ; i++ )
			VFG->view_field_list[i] = (STV_VIEW_FIELD) NULL;

		FREE( VFG->view_field_list );

		VFG->view_field_list = (STV_VIEW_FIELD *) NULL;
	}

	if ( VFG->view_types != NULL )
	{
		for ( i=0 ; i < VFG->nvfields ; i++ )
			VFG->view_types[i] = -1;

		FREE( VFG->view_types );

		VFG->view_types = (int *) NULL;
	}

	if ( VFG->view_storage_orders != NULL )
	{
		for ( i=0 ; i < VFG->nvfields ; i++ )
			VFG->view_storage_orders[i] = -1;

		FREE( VFG->view_storage_orders );

		VFG->view_storage_orders = (int *) NULL;
	}

	if ( VFG->field_values != NULL )
	{
		for ( i=0 ; i < VFG->nvfields ; i++ )
		{
			if ( VFG->field_values[i] != NULL )
			{
				FREE( VFG->field_values[i] );

				VFG->field_values[i] = (STV_VALUE) NULL;
			}
		}

		FREE( VFG->field_values );

		VFG->field_values = (STV_VALUE *) NULL;
	}

	if ( VFG->field_arrayDecls != NULL )
	{
		for ( i=0 ; i < VFG->nvfields ; i++ )
		{
			if ( VFG->field_arrayDecls[i] != NULL )
			{
				FREE( VFG->field_arrayDecls[i] );

				VFG->field_arrayDecls[i] = (int *) NULL;
			}
		}

		FREE( VFG->field_arrayDecls );

		VFG->field_arrayDecls = (int **) NULL;
	}

	VFG->nvfields = -1;

	stv_free_view_particle_list( &(VFG->particle_list) );

	if ( VFG->tid_list != NULL )
	{
		for ( i=0 ; i < VFG->ntids ; i++ )
			VFG->tid_list[i] = -1;

		FREE( VFG->tid_list );

		VFG->tid_list = (int *) NULL;
	}

	VFG->ntids = -1;

	stv_init_region( VFG->visregion );

	stv_init_cell( VFG->viscell );

	VFG->region_time = -1;
	VFG->time_offset = -1;
	VFG->region_cnt = -1;
	VFG->region_id = -1;
	VFG->reply_cnt = -1;
	VFG->frame_ok = -1;
	VFG->ack_cnt = -1;
	VFG->timemax = -1;
	VFG->locked = -1;
	VFG->eff_freq = -1;
	VFG->freq = stvNotTime;
	VFG->flag = -1;
	VFG->bail = -1;

	VFG->next = (STV_VIEW_FIELD_GROUP) NULL;

	FREE( VFG );

	*ptr = (STV_VIEW_FIELD_GROUP) NULL;
}


void
stv_viewer_release_field( ptr, V )
STV_VIEW_FIELD_GROUP *ptr;
STV_VIEWER V;
{
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_FIELD_GROUP vfglast;
	STV_VIEW_FIELD_GROUP vfg;

	if ( ptr == NULL || *ptr == NULL )
		return;

	if ( V != NULL )
	{
		vfglast = (STV_VIEW_FIELD_GROUP) NULL;

		vfg = V->vfglist;

		VFG = *ptr;

		while ( vfg != NULL && vfg != VFG )
		{
			vfglast = vfg;
			vfg = vfg->next;
		}

		if ( vfg != NULL )
		{
			if ( vfglast != NULL )
				vfglast->next = vfg->next;
			
			else
				V->vfglist = vfg->next;
		}
	}

	stv_free_view_field_group( ptr );
}


STV_VIEW_FIELD_GROUP
stv_make_view_field_group( vfields, nvfields )
STV_VIEW_FIELD vfields;
int nvfields;
{
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_FIELD VF;

	int time_scale;
	int max_tids;
	int i, j;
	int flag;
	int vf;

	/* Create New View Field Group Structure */

	VFG = stv_create_view_field_group();

	STV_NO_MEM_TYPE_return( VFG, "stv_make_view_field_group",
		STV_VIEW_FIELD_GROUP );

	/* Count Up View Fields & Max View Field Tasks */

	VFG->nvfields = 0;

	max_tids = 0;

	VF = vfields;

	flag = 0;

	while ( VF != NULL )
	{
		if ( VF->selected )
		{
			if ( !flag++ )
				time_scale = VF->field->time_scale;
			
			else if ( time_scale != VF->field->time_scale )
			{
				sprintf( STVERRTXT,
					"Field Time Scale Mismatch - %d %s vs. %d %s",
					stv_value_of_time( time_scale ),
					stv_time_unit_str( stv_unit_of_time( time_scale ) ),
					stv_value_of_time( VF->field->time_scale ),
					stv_time_unit_str(
						stv_unit_of_time( VF->field->time_scale ) ) );
				STV_ERROR( "stv_make_view_field_group",
						stvStatusBadParam, STVERRTXT );

				FREE( VFG );

				return( (STV_VIEW_FIELD_GROUP) NULL );
			}

			(VFG->nvfields)++;

			max_tids += VF->ntids;
		}

		VF = VF->next;
	}

	/* VFG O.K., Get Next Field Group ID & Region ID */

	VFG->fgid = STV_NEXT_FGID++;

	VFG->region_id = STV_NEXT_REGION_ID++;

	/* Allocate View Fields List */

	VFG->view_field_list = (STV_VIEW_FIELD *) MALLOC(
		(STV_SIZE_T) VFG->nvfields * sizeof( STV_VIEW_FIELD ) );
	STV_MEMCHECK( VFG->view_field_list, "View Field List Ptr",
		return( (STV_VIEW_FIELD_GROUP) NULL ) );

	/* Allocate View Types List */

	VFG->view_types = (int *) MALLOC(
		(STV_SIZE_T) VFG->nvfields * sizeof( int ) );
	STV_MEMCHECK( VFG->view_types, "View Types List Ptr",
		return( (STV_VIEW_FIELD_GROUP) NULL ) );

	/* Allocate View Storage Orders List */

	VFG->view_storage_orders = (int *) MALLOC(
		(STV_SIZE_T) VFG->nvfields * sizeof( int ) );
	STV_MEMCHECK( VFG->view_storage_orders,
		"View Storage Orders List Ptr",
		return( (STV_VIEW_FIELD_GROUP) NULL ) );

	/* Allocate Field Values List */

	VFG->field_values = (STV_VALUE *) MALLOC(
		(STV_SIZE_T) VFG->nvfields * sizeof( STV_VALUE ) );
	STV_MEMCHECK( VFG->field_values, "Field Values List Ptr",
		return( (STV_VIEW_FIELD_GROUP) NULL ) );

	/* Allocate Field Array Decls List */

	VFG->field_arrayDecls = (int **) MALLOC(
		(STV_SIZE_T) VFG->nvfields * sizeof( int * ) );
	STV_MEMCHECK( VFG->field_arrayDecls, "Field Array Decls List Ptr",
		return( (STV_VIEW_FIELD_GROUP) NULL ) );

	/* Initialize Field Lists */

	for ( i=0 ; i < VFG->nvfields ; i++ )
	{
		VFG->field_values[i] = (STV_VALUE) NULL;

		VFG->field_arrayDecls[i] = (int *) NULL;
	}

	/* Initialize Tasks List */

	VFG->tid_list = (int *) MALLOC( (STV_SIZE_T) max_tids
		* sizeof( int ) );
	STV_MEMCHECK( VFG->tid_list, "View Field Group Task List Ptr",
		return( (STV_VIEW_FIELD_GROUP) NULL ) );

	VFG->ntids = 0;

	/* Fill In Tasks */

	VF = vfields;

	vf = 0;

	while ( VF != NULL )
	{
		if ( VF->selected )
		{
			/* Get View Field */

			VFG->view_field_list[ vf ] = VF;

			if ( VF->view_type >= 0 )
				VFG->view_types[ vf ] = VF->view_type;
			
			else
				VFG->view_types[ vf ] = VF->field->type;

			if ( VF->view_storage_order >= 0 )
				VFG->view_storage_orders[ vf ] = VF->view_storage_order;

			else
			{
				VFG->view_storage_orders[ vf ] =
					VF->field->storage_order;
			}

			vf++;

			/* Get View Field Tasks */

			for ( i=0 ; i < VF->ntids ; i++ )
			{
				flag = 0;

				for ( j=0 ; j < VFG->ntids && !flag ; j++ )
				{
					if ( VFG->tid_list[j] == VF->tids[i] )
						flag++;
				}

				if ( !flag )
					VFG->tid_list[ (VFG->ntids)++ ] = VF->tids[i];
			}
		}

		VF = VF->next;
	}

	VFG->bail = stvFalse;

	return( VFG );
}


STV_VIEW_FIELD_GROUP
stv_get_view_field_group_id( vfglist, nvfgs, fgid )
STV_VIEW_FIELD_GROUP vfglist;
int nvfgs;
int fgid;
{
	STV_VIEW_FIELD_GROUP VFG;

	VFG = vfglist;

	while ( VFG != NULL )
	{
		if ( VFG->fgid == fgid )
			return( VFG );

		VFG = VFG->next;
	}

	if ( VFG == NULL )
	{
		if ( STV_STEER_VFG && STV_STEER_VFG->fgid == fgid )
			return( STV_STEER_VFG );

		else if ( STV_SYNC_VFG && STV_SYNC_VFG->fgid == fgid )
			return( STV_SYNC_VFG );
	}

	return( VFG );
}


int
stv_check_view_field_group_task( VFG, tid )
STV_VIEW_FIELD_GROUP VFG;
int tid;
{
	int i;

	for ( i=0 ; i < VFG->ntids ; i++ )
	{
		if ( VFG->tid_list[i] == tid )
			return( stvTrue );
	}

	return( stvFalse );
}


int
stv_check_viewer_notify( V, tid )
STV_VIEWER V;
int tid;
{
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_TASK VT;

	int ret;
	int i;

	ret = stvFalse;

	/* Check View Field Groups */

	VFG = V->vfglist;

	while ( VFG != NULL )
	{
		if ( stv_check_view_field_group_task( VFG, tid ) )
			VFG->bail = stvTrue;

		VFG = VFG->next;
	}

	/* Check Current Viewer Tasks */

	VT = V->vtasks;

	while ( VT != NULL )
	{
		if ( VT->tid == tid )
			ret = stvTrue;

		VT = VT->next;
	}

	return( ret );
}


int
stv_check_view_field_group_notify( V, tid )
STV_VIEWER V;
int tid;
{
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_TASK VT;

	int ret;
	int i;

	ret = stvFalse;

	/* Check View Field Groups */

	VFG = V->vfglist;

	while ( VFG != NULL )
	{
		if ( stv_check_view_field_group_task( VFG, tid ) )
		{
			VFG->bail = stvTrue;

			ret = stvTrue;
		}

		VFG = VFG->next;
	}

	return( ret );
}


STV_VIEW_FIELD
stv_get_view_field_from_group_id( VFG, id, index )
STV_VIEW_FIELD_GROUP VFG;
int id;
int *index;
{
	STV_VIEW_FIELD VF;

	int i;

	for ( i=0 ; i < VFG->nvfields ; i++ )
	{
		VF = VFG->view_field_list[i];

		if ( VF->field->fieldID == id )
		{
			if ( index != NULL )
				*index = i;

			return( VF );
		}
	}

	if ( index != NULL )
		*index = -1;

	return( (STV_VIEW_FIELD) NULL );
}


STV_VIEW_PARAM
stv_create_view_param()
{
	STV_VIEW_PARAM tmp;

	tmp = (STV_VIEW_PARAM) MALLOC(
		sizeof( struct stv_view_param_struct ) );
	STV_MEMCHECK( tmp, "View Parameter Structure",
		return( (STV_VIEW_PARAM) NULL ) );

	tmp->param = (STV_PARAMPTR) NULL;

	tmp->token = -1;

	tmp->changed = -1;

	tmp->tids = (int *) NULL;

	tmp->ntids = -1;

	tmp->next = (STV_VIEW_PARAM) NULL;

	return( tmp );
}


void
stv_free_view_param( ptr )
STV_VIEW_PARAM *ptr;
{
	STV_VIEW_PARAM VP;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VP = *ptr;

	VP->param = (STV_PARAMPTR) NULL;

	VP->token = -1;

	VP->changed = -1;

	if ( VP->tids != NULL )
	{
		FREE( VP->tids );

		VP->tids = (int *) NULL;
	}

	VP->ntids = -1;

	VP->next = (STV_VIEW_PARAM) NULL;

	FREE( VP );

	*ptr = (STV_VIEW_PARAM) NULL;
}


STV_VIEW_PARAM
stv_get_view_param_id( id, vparams, nvparams ) 
int id;
STV_VIEW_PARAM vparams;
int nvparams;
{
	STV_VIEW_PARAM VP;

	VP = vparams;

	while ( VP != NULL )
	{
		if ( VP->param != NULL && VP->param->paramID == id )
			return( VP );

		VP = VP->next;
	}

	return( VP );
}


STV_VIEW_PARAM
stv_get_view_param_name( name, vparams, nvparams )
char *name;
STV_VIEW_PARAM vparams;
int nvparams;
{
	STV_VIEW_PARAM VP;

	VP = vparams;

	while ( VP != NULL )
	{
		if ( VP->param != NULL && !strcmp( VP->param->name, name ) )
			return( VP );

		VP = VP->next;
	}

	return( VP );
}


STV_VIEW_TASK
stv_create_view_task()
{
	STV_VIEW_TASK tmp;

	tmp = (STV_VIEW_TASK) MALLOC(
		sizeof( struct stv_view_task_struct ) );
	STV_MEMCHECK( tmp, "View Task Structure",
		return( (STV_VIEW_TASK) NULL ) );

	tmp->tid = -1;

	tmp->fields = (STV_FIELDPTR *) NULL;

	tmp->nfields = -1;

	tmp->fieldtime = -1;

	tmp->ack = -1;

	tmp->next = (STV_VIEW_TASK) NULL;

	return( tmp );
}


void
stv_free_view_task( ptr )
STV_VIEW_TASK *ptr;
{
	STV_VIEW_TASK VT;

	if ( ptr == NULL || *ptr == NULL )
		return;

	VT = *ptr;

	VT->tid = -1;

	VT->fields = (STV_FIELDPTR *) NULL;

	VT->nfields = -1;

	VT->fieldtime = -1;

	VT->ack = -1;

	VT->next = (STV_VIEW_TASK) NULL;

	FREE( VT );

	*ptr = (STV_VIEW_TASK) NULL;
}


void
stv_free_view_task_list( ptr )
STV_VIEW_TASK *ptr;
{
	STV_VIEW_TASK VTnext;
	STV_VIEW_TASK VT;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	VT = *ptr;

	while ( VT != NULL )
	{
		VTnext = VT->next;

		stv_free_view_task( &VT );

		VT = VTnext;
	}

	*ptr = (STV_VIEW_TASK) NULL;
}


STV_VIEW_TASK
stv_get_view_task_tid( tid, vtasks, nvtasks )
int tid;
STV_VIEW_TASK vtasks;
int nvtasks;
{
	STV_VIEW_TASK VT;

	VT = vtasks;

	while ( VT != NULL )
	{
		if ( VT->tid == tid )
			return( VT );

		VT = VT->next;
	}

	return( VT );
}


STV_VIEWER
stv_create_viewer()
{
	STV_VIEWER tmp;

	tmp = (STV_VIEWER) MALLOC( sizeof( struct stv_viewer_struct ) );
	STV_MEMCHECK( tmp, "Viewer Structure",
		return( (STV_VIEWER) NULL ) );

	tmp->vfglist = (STV_VIEW_FIELD_GROUP) NULL;
	tmp->nvfgs = -1;

	tmp->vfields = (STV_VIEW_FIELD) NULL;
	tmp->nvfields = -1;

	tmp->vparams = (STV_VIEW_PARAM) NULL;
	tmp->nvparams = -1;

	tmp->vtasks = (STV_VIEW_TASK) NULL;
	tmp->nvtasks = -1;

	tmp->particles = (STV_PARTICLEPTR) NULL;
	tmp->nparticles = -1;

	tmp->decomps = (STV_DECOMPPTR) NULL;
	tmp->ndecomps = -1;

	tmp->fields = (STV_FIELDPTR) NULL;
	tmp->nfields = -1;

	tmp->params = (STV_PARAMPTR) NULL;
	tmp->nparams = -1;

	tmp->init_seq = -1;
	tmp->init_gen = -1;

	tmp->msgtag = -1;
	tmp->ntasks = -1;
	tmp->id = -1;

	tmp->next = (STV_VIEWER) NULL;

	return( tmp );
}


void
stv_free_viewer( ptr )
STV_VIEWER *ptr;
{
	STV_VIEWER Vptr;
	STV_VIEWER V;

	if ( ptr == NULL || *ptr == NULL )
		return;

	V = *ptr;

	/* Reset Main Fields */

	stv_viewer_reset( V );

	/* Invalidate Counters */

	V->nvfgs = -1;

	V->nvfields = -1;
	V->nvparams = -1;
	V->nvtasks = -1;

	V->nparticles = -1;
	V->ndecomps = -1;
	V->nfields = -1;
	V->nparams = -1;

	/* Find & Dereference Viewer from Master List */

	if ( V == STVVIEWERS )
		STVVIEWERS = V->next;
	
	else
	{
		Vptr = STVVIEWERS;

		while ( Vptr != NULL && Vptr->next != V )
			Vptr = Vptr->next;
	
		if ( Vptr != NULL )
			Vptr->next = V->next;
	}

	V->next = (STV_VIEWER) NULL;

	nSTVVIEWERS--;

	/* Free Overall Structure */

	FREE( V );

	*ptr = (STV_VIEWER) NULL;
}


void
stv_viewer_reset( V )
STV_VIEWER V;
{
	STV_VIEW_FIELD_GROUP VFGnext;
	STV_VIEW_FIELD_GROUP VFG;

	STV_VIEW_FIELD VFnext;
	STV_VIEW_FIELD VF;

	STV_VIEW_PARAM VPnext;
	STV_VIEW_PARAM VP;

	STV_VIEW_TASK VTnext;
	STV_VIEW_TASK VT;

	STV_PARTICLEPTR PTnext;
	STV_PARTICLEPTR PT;

	STV_DECOMPPTR Dnext;
	STV_DECOMPPTR D;

	STV_FIELDPTR Fnext;
	STV_FIELDPTR F;

	STV_PARAMPTR Pnext;
	STV_PARAMPTR P;

	/* Free Any Viewer Field Groups */

	if ( V->vfglist != NULL )
	{
		VFG = V->vfglist;

		while ( VFG != NULL )
		{
			VFGnext = VFG->next;

			stv_free_view_field_group( &VFG );

			VFG = VFGnext;
		}

		V->vfglist = (STV_VIEW_FIELD_GROUP) NULL;
	}

	V->nvfgs = 0;

	/* Free Any Viewer Fields */

	if ( V->vfields != NULL )
	{
		VF = V->vfields;

		while ( VF != NULL )
		{
			VFnext = VF->next;

			stv_free_view_field( &VF );

			VF = VFnext;
		}

		V->vfields = (STV_VIEW_FIELD) NULL;
	}

	V->nvfields = 0;

	/* Free Any Viewer Parameters */

	if ( V->vparams != NULL )
	{
		VP = V->vparams;

		while ( VP != NULL )
		{
			VPnext = VP->next;

			stv_free_view_param( &VP );

			VP = VPnext;
		}

		V->vparams = (STV_VIEW_PARAM) NULL;
	}

	V->nvparams = 0;

	/* Free Any Viewer Tasks */

	stv_free_view_task_list( &(V->vtasks) );

	V->nvtasks = 0;

	/* Free Any Decompositions */

	if ( V->decomps != NULL )
	{
		D = V->decomps;

		while ( D != NULL )
		{
			Dnext = D->next;

			stv_free_decomp( &D );

			D = Dnext;
		}

		V->decomps = (STV_DECOMPPTR) NULL;
	}

	V->ndecomps = 0;

	/* Free Any Particles */

	if ( V->particles != NULL )
	{
		PT = V->particles;

		while ( PT != NULL )
		{
			PTnext = PT->next;

			stv_free_particle( &PT );

			PT = PTnext;
		}

		V->particles = (STV_PARTICLEPTR) NULL;
	}

	V->nparticles = 0;

	/* Free Any Fields */

	if ( V->fields != NULL )
	{
		F = V->fields;

		while ( F != NULL )
		{
			Fnext = F->next;

			stv_free_field( &F );

			F = Fnext;
		}

		V->fields = (STV_FIELDPTR) NULL;
	}

	V->nfields = 0;

	/* Free Any Parameters */

	if ( V->params != NULL )
	{
		P = V->params;

		while ( P != NULL )
		{
			Pnext = P->next;

			stv_free_param( &P );

			P = Pnext;
		}

		V->params = (STV_PARAMPTR) NULL;
	}

	V->nparams = 0;

	/* Reset Other Values */

	V->init_seq = -1;
	V->init_gen = -1;

	V->msgtag = -1;

	V->ntasks = -1;

	V->id = -1;
}


STV_VIEWER
stv_find_viewer( vid )
int vid;
{
	STV_VIEWER V;

	V = STVVIEWERS;

	while ( V != NULL )
	{
		if ( V->id == vid )
			return( V );

		V = V->next;
	}

	return( (STV_VIEWER) NULL );
}


STV_PACK_INDEX
stv_create_pack_index()
{
	STV_PACK_INDEX tmp;

	tmp = (STV_PACK_INDEX) MALLOC(
		sizeof( struct stv_pack_index_struct ) );
	STV_MEMCHECK( tmp, "Pack Index Structure",
		return( (STV_PACK_INDEX) NULL ) );

	tmp->gstart = -1;
	tmp->gstride = -1;
	tmp->lstart = -1;
	tmp->lstride = -1;
	tmp->max = -1;

	tmp->next = (STV_PACK_INDEX) NULL;

	return( tmp );
}


void
stv_free_pack_index( ptr )
STV_PACK_INDEX *ptr;
{
	STV_PACK_INDEX PX;

	if ( ptr == NULL || *ptr == NULL )
		return;

	PX = *ptr;

	PX->gstart = -1;
	PX->gstride = -1;
	PX->lstart = -1;
	PX->lstride = -1;
	PX->max = -1;

	PX->next = (STV_PACK_INDEX) NULL;

	FREE( PX );

	*ptr = (STV_PACK_INDEX) NULL;
}


void
stv_free_pack_index_list( ptr )
STV_PACK_INDEX *ptr;
{
	STV_PACK_INDEX PXnext;
	STV_PACK_INDEX PX;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	PX = *ptr;

	while ( PX != NULL )
	{
		PXnext = PX->next;

		stv_free_pack_index( &PX );

		PX = PXnext;
	}

	*ptr = (STV_PACK_INDEX) NULL;
}


STV_PACK_DIM
stv_create_pack_dim()
{
	STV_PACK_DIM tmp;

	tmp = (STV_PACK_DIM) MALLOC( sizeof( struct stv_pack_dim_struct ) );
	STV_MEMCHECK( tmp, "Pack Dimension Structure",
		return( (STV_PACK_DIM) NULL ) );

	tmp->pindices = (STV_PACK_INDEX) NULL;

	tmp->current = (STV_PACK_INDEX) NULL;

	tmp->gindex = -1;
	tmp->lindex = -1;

	tmp->type = -1;

	tmp->dim = -1;

	tmp->next = (STV_PACK_DIM) NULL;

	return( tmp );
}


void
stv_free_pack_dim( ptr )
STV_PACK_DIM *ptr;
{
	STV_PACK_DIM PD;

	if ( ptr == NULL || *ptr == NULL )
		return;

	PD = *ptr;

	PD->pindices = (STV_PACK_INDEX) NULL;

	PD->current = (STV_PACK_INDEX) NULL;

	PD->gindex = -1;
	PD->lindex = -1;

	PD->type = -1;

	PD->dim = -1;

	PD->next = (STV_PACK_DIM) NULL;

	FREE( PD );

	*ptr = (STV_PACK_DIM) NULL;
}


void
stv_free_pack_dim_list( ptr )
STV_PACK_DIM *ptr;
{
	STV_PACK_DIM PDnext;
	STV_PACK_DIM PD;

	if ( ptr == NULL || *ptr == NULL )
		return;
	
	PD = *ptr;

	while ( PD )
	{
		PDnext = PD->next;

		stv_free_pack_dim( &PD );

		PD = PDnext;
	}

	*ptr = (STV_PACK_DIM) NULL;
}


void
stv_check_decomp_axis_info( axisType, axisInfo,
	axisLBounds, axisUBounds, nbounds, glb, gub, ntasks )
int axisType;
int *axisInfo;
int *axisLBounds;
int *axisUBounds;
int nbounds;
int glb, gub;
int ntasks;
{
	int size;
	int b;

	if ( axisType == stvBlock )
	{
		if ( *axisInfo == 0 || *axisInfo < -1 )
		{
			sprintf( STVERRTXT, "Warning:  Invalid Block Size of %d",
				*axisInfo );

			STV_ERROR( "stv_check_decomp_axis_info", stvStatusBadParam,
				STVERRTXT );

			*axisInfo = -1;
		}

		else if ( *axisInfo == -1 )
		{
			size = gub - glb + 1;

			*axisInfo = (int) ceil( (double) size / (double) ntasks );
		}
	}

	else if ( axisType == stvCyclic )
	{
		if ( *axisInfo == 0 || *axisInfo < -1 )
		{
			sprintf( STVERRTXT, "Warning:  Invalid Cycle Size of %d",
				*axisInfo );

			STV_ERROR( "stv_check_decomp_axis_info", stvStatusBadParam,
				STVERRTXT );

			*axisInfo = -1;
		}

		else if ( *axisInfo == -1 )
			*axisInfo = 1;
	}

	else if ( axisType == stvExplicit )
	{
		for ( b=0 ; b < nbounds ; b++ )
		{
			if ( axisLBounds[b] < glb )
				axisLBounds[b] = glb;

			else if ( axisLBounds[b] > gub )
				axisLBounds[b] = axisUBounds[b] = gub;

			if ( axisUBounds[b] > gub )
				axisUBounds[b] = gub;
		
			else if ( axisUBounds[b] < glb )
				axisLBounds[b] = axisUBounds[b] = glb;
		}
	}

	else if ( axisType != stvCollapse )
	{
		sprintf( STVERRTXT, "Unknown Decomposition Axis Type = %d",
			axisType );

 		STV_ERROR( "stv_check_decomp_axis_info", stvStatusBadParam,
				STVERRTXT );

		*axisInfo = -1;

		for ( b=0 ; b < nbounds ; b++ )
		{
			axisLBounds[b] = -1;
			axisUBounds[b] = -1;
		}
	}
}


int
stv_get_proc_dim_size( axisType, axisInfo,
	axisLBounds, axisUBounds, nbounds, size, p, ntasks )
int axisType;
int axisInfo;
int *axisLBounds;
int *axisUBounds;
int nbounds;
int size;
int p;
int ntasks;
{
	int num_add_whole_cycles;
	int num_add_part_cycles;
	int min_elems_per_proc;
	int num_full_procs;
	int dsize;
	int left;
	int b;

	if ( size <= 0 )
		dsize = 0;

	else if ( axisType == stvBlock )
	{
		if ( p < size / axisInfo )
			dsize = axisInfo;

		else if ( p == size / axisInfo )
			dsize = size % axisInfo;

		else
			dsize = 0;
	}

	else if ( axisType == stvCyclic )
	{
		/* Add in Base Cycle Distribution */

		min_elems_per_proc = ( size / ( axisInfo * ntasks ) )
			* axisInfo;

		if ( min_elems_per_proc > 0 )
		{
			dsize = min_elems_per_proc;

			left = size - ( min_elems_per_proc * ntasks );

			num_add_whole_cycles = left / axisInfo;

			num_add_part_cycles = left % axisInfo;
		}

		else
		{
			dsize = 0;

			num_add_whole_cycles = size / axisInfo;

			num_add_part_cycles = size % axisInfo;
		}

		/* Add in Any Additional Distribution */

		if ( p < num_add_whole_cycles )
			dsize += axisInfo;
		
		else if ( p == num_add_whole_cycles )
			dsize += num_add_part_cycles;
	}

	else if ( axisType == stvExplicit )
	{
		dsize = 0;

		for ( b=0 ; b < nbounds ; b++ )
			dsize += axisUBounds[b] - axisLBounds[b] + 1;
	}

	else if ( axisType == stvCollapse )
		dsize = size;

	else
	{
		sprintf( STVERRTXT, "Unknown Decomposition Axis Type = %d",
			axisType );

		STV_ERROR( "stv_get_proc_dim_size", stvStatusBadParam,
			STVERRTXT );

		dsize = stvStatusBadParam;
	}

	return( dsize );
}


int *
stv_add_to_list( list, size, elem )
int *list;
int size;
int elem;
{
	int *tmp;

	int i;

	tmp = (int *) MALLOC( (STV_SIZE_T) (size + 1) * sizeof( int ) );
	STV_MEMCHECK( tmp, "New Integer List Pointer",
		return( (int *) NULL ) );

	for ( i=0 ; i < size ; i++ )
		tmp[i] = list[i];
	
	tmp[ size ] = elem;

	if ( list != NULL )
		FREE( list );

	return( tmp );
}


/* Dump Out VFG Field Frame Array */

void
stv_dump_vfg_field_array( VFG, index )
STV_VIEW_FIELD_GROUP VFG;
int index;
{
	STV_VIEW_FIELD VF;

	VF = VFG->view_field_list[ index ];

	if ( VF->field->decomp == NULL )
	{
		STV_ERROR( "stv_dump_vfg_field_array", stvStatusBadParam,
			"Not a Contiguous Decomposition." );

		return;
	}

	stv_dump_field_array(
		VFG->field_values[ index ],
		VFG->field_arrayDecls[ index ],
		VF->field->decomp->dataDim,
		VFG->visregion,
		VFG->viscell,
		VFG->view_types[ index ],
		VFG->view_storage_orders[ index ] );
}


/* Dump Out User Field Array */

void
stv_dump_user_field_array( var, array_decl, dim, offsets,
	type, storage_order )
STV_VALUE var;
int *array_decl;
int dim;
int *offsets;
int type;
int storage_order;
{
	STV_REGION R;

	STV_CELL C;

	int i;

	stv_init_region( R );

	stv_init_cell( C );

	for ( i=0 ; i < dim ; i++ )
	{
		STV_REGION_MIN( R, i ) = offsets[i];
		STV_REGION_MAX( R, i ) = offsets[i] + array_decl[i] - 1;

		STV_CELL_OF_DIM( C, i ) = 1;
	}

	stv_dump_field_array( var, array_decl, dim, R, C,
		type, storage_order );
}


/* Dump Out Generic Field Array */

void
stv_dump_field_array( var, array_decl, dim, R, C, type, storage_order )
STV_VALUE var;
int *array_decl;
int dim;
STV_REGION R;
STV_CELL C;
int type;
int storage_order;
{
	STV_VALUE ptr;

	char *str;

	int indices[STV_MAX_DIM];

	int index;
	int i;

	/* Dump Storage Order */

	printf( "(%s Data Storage Order)\n\n",
		stv_storage_order_str( storage_order ) );

	/* Initialize Local Index */

	for ( i=0 ; i < dim ; i++ )
		indices[i] = 0;

	/* Dump Out Elements */

	stv_dump_base_indices( indices, dim, R, C, storage_order );

	ptr = var;

	do
	{
		str = stv_str_of_typed_data( ptr, type );

		printf( "%s ", str );

		FREE( str );

		ptr = stv_get_typed_ptr_offset( ptr, 1, type );
	}
	while ( ptr != NULL
		&& stv_incr_field_index( dim, indices, array_decl,
			R, C, storage_order, stvTrue ) );

	printf( "\n" );
}


void
stv_dump_base_indices( indices, dim, R, C, storage_order )
int *indices;
int dim;
STV_REGION R;
STV_CELL C;
int storage_order;
{
	int skip1, skip2;
	int i;

	if ( storage_order == stvColumnMajor )
	{
		skip1 = 0;
		skip2 = 1;
	}

	else if ( storage_order == stvRowMajor )
	{
		skip1 = dim - 2;
		skip2 = dim - 1;
	}

	else
	{
		sprintf( STVERRTXT, "Unknown Data Storage Order, ID = %d",
			storage_order );

		STV_ERROR( "stv_dump_base_indices", stvStatusBadParam,
			STVERRTXT );

		return;
	}

	printf( "( " );

	for ( i=0 ; i < dim ; i++ )
	{
		if ( i != skip1 && i != skip2 )
		{
			printf( "%d%s",
				STV_REGION_MIN( R, i )
					+ ( STV_CELL_OF_DIM( C, i ) * indices[i] ),
				( i < dim - 1 ) ? ", " : "" );
		}
		
		else
		{
			printf( "%d to %d by %d%s",
				STV_REGION_MIN( R, i ),
				STV_REGION_MAX( R, i ),
				STV_CELL_OF_DIM( C, i ),
				( i < dim - 1 ) ? ", " : "" );
		}
	}

	printf( " ):\n" );
}


/* Increment Indices for Field Array Dumping */

int
stv_incr_field_index( dim, indices, max, R, C, storage_order, nlflag )
int dim;
int *indices;
int *max;
STV_REGION R;
STV_CELL C;
int storage_order;
int nlflag;
{
	int dump_base;
	int index;
	int flag;
	int incr;

	if ( storage_order == stvColumnMajor )
	{
		index = 0;
		incr = 1;
	}

	else if ( storage_order == stvRowMajor )
	{
		index = dim - 1;
		incr = -1;
	}

	else
	{
		sprintf( STVERRTXT, "Unknown Data Storage Order, ID = %d",
			storage_order );

		STV_ERROR( "stv_incr_field_index", stvStatusBadParam,
			STVERRTXT );

		return( stvFalse );
	}

	do
	{
		flag = 0;

		(indices[index])++;

		if ( indices[index] >= max[index] )
		{
			indices[index] = 0;

			flag++;
		}

		if ( flag )
		{
			index += incr;

			if ( index >= dim || index < 0 )
			{
				if ( nlflag )
					printf( "\n" );

				return( stvFalse );
			}
		}
	}
	while ( flag );

	if ( nlflag )
	{
		if ( ( incr > 0 && index > 1 )
			|| ( incr < 0 && index < dim - 2 ) )
		{
			dump_base = stvTrue;
		}

		else
			dump_base = stvFalse;

		while ( ( incr > 0 && index > 0 )
			|| ( incr < 0 && index < dim - 1 ) )
		{
			printf( "\n" );

			index -= incr;
		}

		if ( dump_base )
			stv_dump_base_indices( indices, dim, R, C, storage_order );
	}

	return( stvTrue );
}


/* Region & Cell Size Routines */


void
stv_init_region( R )
STV_REGION R;
{
	int i;

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
	{
		STV_REGION_MIN( R, i ) = -1;
		STV_REGION_MAX( R, i ) = -1;
	}
}


int
stv_match_region( R1, R2, dim )
STV_REGION R1, R2;
int dim;
{
	int i;

	for ( i=0 ; i < dim ; i++ )
	{
		if ( STV_REGION_MIN( R1, i ) != STV_REGION_MIN( R2, i ) )
			return( stvFalse );

		if ( STV_REGION_MAX( R1, i ) != STV_REGION_MAX( R2, i ) )
			return( stvFalse );
	}

	return( stvTrue );
}


void
stv_copy_region( R1, R2 )
STV_REGION R1, R2;
{
	int i;

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
	{
		STV_REGION_MIN( R1, i ) = STV_REGION_MIN( R2, i );

		STV_REGION_MAX( R1, i ) = STV_REGION_MAX( R2, i );
	}
}


void
stv_init_cell( C )
STV_CELL C;
{
	int i;

	for ( i=0 ; i < STV_CELL_SIZE( STV_MAX_DIM ) ; i++ )
		STV_CELL_OF_DIM( C, i ) = -1;
}


int
stv_match_cell( C1, C2, dim )
STV_CELL C1, C2;
int dim;
{
	int i;

	for ( i=0 ; i < dim ; i++ )
	{
		if ( C1[i] != C2[i] )
			return( stvFalse );
	}

	return( stvTrue );
}


void
stv_copy_cell( C1, C2 )
STV_CELL C1, C2;
{
	int i;

	for ( i=0 ; i < STV_CELL_SIZE( STV_MAX_DIM ) ; i++ )
		C1[i] = C2[i];
}


double
stv_fabs( dval )
double dval;
{
	return( fabs( dval ) );
}


int
stv_verify_system_contexts( insertflag, vflag )
int insertflag;
int vflag;
{
	int new = 0;
	int cc;

	/* Verify System Viewer Context */

	if ( STV_VIEWER_CTX == (STV_CONTEXT) -1 )
	{
		if ( (cc = stv_verify_context( "Viewer", stvViewerContext,
				&STV_VIEWER_CTX, insertflag, vflag )) < 0 )
			return( cc );

		new++;
	}

	/* Verify System Cpd Context */

	if ( STV_CPD_CTX == (STV_CONTEXT) -1 )
	{
		if ( (cc = stv_verify_context( "Cpd", stvCpdContext,
				&STV_CPD_CTX, insertflag, vflag )) < 0 )
			return( cc );

		new++;
	}

	/* Verify System Application Context */

	if ( STV_APP_CTX == (STV_CONTEXT) -1 )
	{
		if ( (cc = stv_verify_context( "Application", stvAppContext,
				&STV_APP_CTX, insertflag, vflag )) < 0 )
			return( cc );

		new++;
	}

	return( ( new ) ? new : stvStatusOk );
}


int
stv_verify_context( desc, name, ctx, insertflag, vflag )
char *desc;
char *name;
STV_CONTEXT *ctx;
int insertflag;
int vflag;
{
	STV_CONTEXT tmpctx;

	int index;
	int cc;

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_VIEWER | STV_DEBUG_ULIB | STV_DEBUG_CPD ) )
	{
		printf( "stv_verify_context(): Verifying %s Context (%s).\n",
				desc, name );
		fflush( stdout );
	}

	/* Check for Context */

	cc = stv_name_lookup_ctx( name, 0, &tmpctx, &index );

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_VIEWER | STV_DEBUG_ULIB | STV_DEBUG_CPD ) )
	{
		printf( "%s: got tmpctx=%lx index=%d cc=%d.\n",
				"stv_verify_context()", (long) tmpctx, index, cc );
		fflush( stdout );
	}

	if ( !cc || index < 0 )
	{
		/* Insert New Context */

		if ( insertflag )
		{
			stv_create_context( &tmpctx );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_ULIB
							| STV_DEBUG_CPD ) )
			{
				printf( "%s: attempting insert of new context %lx.\n",
						"stv_verify_context()", (long) tmpctx );
				fflush( stdout );
			}

			cc = stv_name_insert_ctx( name, 0, &tmpctx, &index );

			if ( STV_OPT_VERBOSE
					& ( STV_DEBUG_VIEWER | STV_DEBUG_ULIB
							| STV_DEBUG_CPD ) )
			{
				printf( "%s: got back index=%d cc=%d.\n",
						"stv_verify_context()", index, cc );
				fflush( stdout );
			}

			/* Error, or Someone Else Already Inserted */

			if ( !cc || index < 0 )
			{
				if ( STV_OPT_VERBOSE
						& ( STV_DEBUG_VIEWER | STV_DEBUG_ULIB
								| STV_DEBUG_CPD ) )
				{
					printf( "%s: insert attempt failed, re-lookup...\n",
							"stv_verify_context()" );
					fflush( stdout );
				}

				/* Free the Context We Made */

				stv_free_context( &tmpctx );

				/* See if Context is There Now */

				cc = stv_name_lookup_ctx( name, 0, &tmpctx, &index );

				if ( !cc || index < 0 )
				{
					sprintf( STVERRTXT,
							"Error Inserting System %s Context (%s)",
							desc, name );
					STV_ERROR( "stv_verify_context", stvStatusBadParam,
							STVERRTXT );
					return( stvStatusBadParam );
				}

				else
					*ctx = tmpctx;
			}

			else
				*ctx = tmpctx;
		}

		else
		{
			if ( vflag )
			{
				sprintf( STVERRTXT,
						"Warning: System %s Context (%s) Not Found",
						desc, name );
				STV_ERROR( "stv_verify_context", stvStatusBadParam,
						STVERRTXT );
			}

			return( stvStatusBadParam );
		}
	}

	else
		*ctx = tmpctx;

	if ( STV_OPT_VERBOSE
			& ( STV_DEBUG_VIEWER | STV_DEBUG_ULIB | STV_DEBUG_CPD ) )
	{
		printf( "%s: Returning %s Context as %lx.\n",
				"stv_verify_context()", desc, (long) *ctx );
		fflush( stdout );
	}

	return( stvStatusOk );
}


int
stv_normalize_time_of( value, units )
int value;
int units;
{
	TIME_CONVERSION *TC;

	int val;
	int t;

	if ( (unsigned) value <= (unsigned) stvTimeMaxValue )
	{
		sprintf( STVERRTXT,
			"Normalizing Unnecessary:  %u %s...",
			value, stv_time_unit_str( units ) );
		STV_ERROR( "stv_normalize_time_of", stvStatusBadParam,
			STVERRTXT );

		return( stv_time_of( value, units ) );
	}

	TC = &(STV_TIME_CONV[0]);

	while ( TC->units && TC->units != units )
		TC++;

	if ( TC->units )
		val = (unsigned) value / (unsigned) (TC->value);

	while ( TC->units && (unsigned) val > (unsigned) stvTimeMaxValue )
	{
		TC++;

		if ( TC->units )
			val /= TC->value;
	}

	if ( TC->units )
	{
		if ( !val )
			val = 1;

		t = stv_time_of( val, TC->conv_units );

		sprintf( STVERRTXT,
			"Time Unit Overflow: Approximated %u %s -> %u %s...",
			value, stv_time_unit_str( units ),
			val, stv_time_unit_str( TC->conv_units ) );
		STV_ERROR( "stv_normalize_time_of", stvStatusBadParam,
			STVERRTXT );
	}

	else
	{
		t = stv_time_of( stvTimeMaxValue, units );

		sprintf( STVERRTXT,
			"Time Unit Overflow: Truncated %u %s at %u %s...",
			value, stv_time_unit_str( units ),
			stvTimeMaxValue, stv_time_unit_str( units ) );
		STV_ERROR( "stv_normalize_time_of", stvStatusBadParam,
			STVERRTXT );
	}

	return( t );
}


int
stv_time_of_double( dvalue, units )
double dvalue;
int units;
{
	TIME_CONVERSION *TCstart;
	TIME_CONVERSION *TC;

	int ivalue;
	int done;

	TC = TCstart = &(STV_TIME_CONV[0]);

	while ( TC->units && TC->conv_units != units )
		TC++;

	ivalue = (int) dvalue;

	done = 0;

	while ( !done
			&& !STV_EQUIV( dvalue, (double) ivalue ) )
	{
		dvalue *= (float) TC->value;

		ivalue = (int) dvalue;

		units = TC->units;

		if ( TC != TCstart )
			TC--;
		
		else
			done++;
	}

	return( stv_time_of( ivalue, units ) );
}


int
stv_reconcile_times( value1, units1, value2, units2 )
int *value1;
int *units1;
int *value2;
int *units2;
{
	TIME_CONVERSION *TC;

	/* Verify Args */

	if ( value1 == NULL || units1 == NULL
			|| value2 == NULL || units2 == NULL )
	{
		STV_ERROR( "stv_reconcile_times", stvStatusBadParam,
			"NULL pointers to time values / units!" );
		
		return( stvFalse );
	}

	/* Check Units */

	if ( (*units1) == (*units2) )
		return( stvTrue );

	/* Convert Times */

	TC = &(STV_TIME_CONV[0]);

	if ( (*units2) > (*units1) )
	{
		while ( TC->units && TC->conv_units != (*units2) )
			TC++;

		if ( TC->units )
		{
			while ( TC->units && TC->conv_units != (*units1) )
			{
				(*value2) *= TC->value;

				TC--;
			}

			(*units2) = (*units1);
		}

		else
		{
			sprintf( STVERRTXT,
				"Unknown Time Unit Conversion: %d %s (%d) - %s",
				(*value2), stv_time_unit_str( (*units2) ), (*units2),
				"Time Unchanged" );
			STV_ERROR( "stv_reconcile_times", stvStatusBadParam,
				STVERRTXT );

			return( stvFalse );
		}
	}

	else
	{
		while ( TC->units && TC->conv_units != (*units1) )
			TC++;

		if ( TC->units )
		{
			while ( TC->units && TC->conv_units != (*units2) )
			{
				(*value1) *= TC->value;

				TC--;
			}

			(*units1) = (*units2);
		}

		else
		{
			sprintf( STVERRTXT,
				"Unknown Time Unit Conversion: %d %s (%d) - %s",
				(*value1), stv_time_unit_str( (*units1) ), (*units1),
				"Time Corrupted" );
			STV_ERROR( "stv_reconcile_times", stvStatusBadParam,
				STVERRTXT );

			return( stvFalse );
		}
	}

	return( stvTrue );
}


void
stv_incr_time( base, units, tval, tunits )
int *base;
int *units;
int tval;
int tunits;
{
	TIME_CONVERSION *TC;

	if ( base == NULL || units == NULL )
	{
		STV_ERROR( "stv_incr_time", stvStatusBadParam,
			"NULL pointers to base time value / units!" );
		return;
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_TIME )
	{
		printf( "%s:  base time = %d %s, incr by %d %s\n",
			"stv_incr_time()", *base, stv_time_unit_str( *units ),
			tval, stv_time_unit_str( tunits ) );
		fflush( stdout );
	}

	if ( !stv_reconcile_times( base, units, &tval, &tunits ) )
	{
		STV_ERROR( "stv_incr_time", stvStatusBadParam,
			"Could Not Reconcile Base Time and Increment Time" );
		return;
	}

	*base += tval;

	if ( STV_OPT_VERBOSE & STV_DEBUG_TIME )
	{
		printf( "\tnew base time = %d %s\n",
			*base, stv_time_unit_str( *units ) );
		fflush( stdout );
	}
}


float
stv_time_ratio( tv1, tv2 )
int tv1;
int tv2;
{
	TIME_CONVERSION *TC;

	float ratio;

	int value1;
	int value2;

	int units1;
	int units2;

	if ( !stv_is_time_value( tv1 ) || !stv_is_time_value( tv2 ) )
		return( 1.0 );

	value1 = stv_value_of_time( tv1 );
	value2 = stv_value_of_time( tv2 );

	units1 = stv_unit_of_time( tv1 );
	units2 = stv_unit_of_time( tv2 );

	if ( !stv_reconcile_times( &value1, &units1, &value2, &units2 ) )
	{
		STV_ERROR( "stv_time_ratio", stvStatusBadParam,
			"Could Not Reconcile Times" );
		return( 1.0 );
	}

	ratio = (float) value1 / (float) value2;

	return( ratio );
}


void
stv_bump_field_times( fields, nfields )
STV_FIELDPTR fields;
int nfields;
{
	STV_FIELDPTR F;

	int tunits;
	int tval;

	F = fields;

	while ( F != NULL )
	{
		if ( F->data_ready )
		{
			tval = stv_value_of_time( F->time_scale );
			tunits = stv_unit_of_time( F->time_scale );

			stv_incr_time( &(F->time_base), &(F->time_units),
				tval, tunits );
		}

		F = F->next;
	}
}


void
stv_dump_debug_flag( flag, nlflag )
int flag;
int nlflag;
{
	unsigned int bit;

	int first = 1;

	bit = 0x1;

	while ( bit < STV_DEBUG_ALL )
	{
		if ( bit & flag )
			printf( "%s%s",
				( first ) ? ( first = 0, "" ) : " ",
				stv_debug_flag_str( bit ) );
		
		bit <<= 1;
	}

	if ( nlflag )
		printf( "\n" );
}


char *
stv_debug_flag_str( flag )
int flag;
{
	char *tmp;

	switch ( flag )
	{
		case STV_DEBUG_MSG:			tmp = "MSG";			break;
		case STV_DEBUG_SPAWN:		tmp = "SPAWN";			break;
		case STV_DEBUG_ULIB:		tmp = "ULIB";			break;
		case STV_DEBUG_FLUSH:		tmp = "FLUSH";			break;
		case STV_DEBUG_NOTIFY:		tmp = "NOTIFY";			break;
		case STV_DEBUG_MEM:			tmp = "MEM";			break;
		case STV_DEBUG_TIME:		tmp = "TIME";			break;
		case STV_DEBUG_FREQ:		tmp = "FREQ";			break;
		case STV_DEBUG_PACK:		tmp = "PACK";			break;
		case STV_DEBUG_PACK_INDEX:	tmp = "PACK_INDEX";		break;
		case STV_DEBUG_VIEWER:		tmp = "VIEWER";			break;
		case STV_DEBUG_STEER:		tmp = "STEER";			break;
		case STV_DEBUG_APPINFO:		tmp = "APPINFO";		break;
		case STV_DEBUG_CPD:			tmp = "CPD";			break;
		case STV_DEBUG_RESTART:		tmp = "RESTART";		break;
		case STV_DEBUG_REGION:		tmp = "REGION";			break;
		case STV_DEBUG_CKPT:		tmp = "CKPT";			break;
		case STV_DEBUG_CKPT_DATA:	tmp = "CKPT_DATA";		break;

		default:
			tmp = "UnknownFlag";
			break;
	}

	return( tmp );
}


char *
stv_msg_op_str( id )
int id;
{
	char *tmp;

	switch ( id )
	{
		case stvInit0:			tmp = "stvInit0";				break;
		case stvInit1:			tmp = "stvInit1";				break;

		case stvFieldRqst:		tmp = "stvFieldRqst";			break;
		case stvFieldTime:		tmp = "stvFieldTime";			break;
		case stvFieldTstart:	tmp = "stvFieldTstart";			break;
		case stvFieldAck:		tmp = "stvFieldAck";			break;
		case stvFieldHaltGroup:	tmp = "stvFieldHaltGroup";		break;
		case stvFieldHaltAll:	tmp = "stvFieldHaltAll";		break;

		case stvXon:			tmp = "stvXon";					break;
		case stvRegion:			tmp = "stvRegion";				break;
		case stvVisRegion:		tmp = "stvVisRegion";			break;
		case stvVisFrequency:	tmp = "stvVisFrequency";		break;
		case stvNewParams:		tmp = "stvNewParams";			break;
		case stvTimeSync:		tmp = "stvTimeSync";			break;
		case stvAppLock:		tmp = "stvAppLock";				break;
		case stvAppUnlock:		tmp = "stvAppUnlock";			break;
		case stvSetopt:			tmp = "stvSetopt";				break;

		case stvNotify:			tmp = "stvNotify";				break;
		case stvReconfigure:	tmp = "stvReconfigure";			break;
		case stvDisconnect:		tmp = "stvDisconnect";			break;

		case stvCheckpointDataSubmit:
								tmp = "stvCheckpointDataSubmit";
																break;
		case stvCheckpointDataUpdate:
								tmp = "stvCheckpointDataUpdate";
																break;
		case stvCpdHalt:		tmp = "stvCpdHalt";				break;
		case stvRestart:		tmp = "stvRestart";				break;
		case stvConfigCpd:		tmp = "stvConfigCpd";			break;
		case stvCheckpointForward:
								tmp = "stvCheckpointForward";	break;
		case stvRequestWhole:	tmp = "stvRequestWhole";		break;
		case stvRequestCheckpoint:
								tmp = "stvRequestCheckpoint";	break;
		case stvWholeCheckpoint:
								tmp = "stvWholeCheckpoint";		break;
		case stvCpdXon:			tmp = "stvCpdXon";				break;
		case stvCpdStabilizeRqst:
								tmp = "stvCpdStabilizeRqst";	break;
		case stvCpdStabilizeReply:
								tmp = "stvCpdStabilizeReply";	break;
		case stvCpdCheckpointStable:
								tmp = "stvCpdCheckpointStable";	break;
		case stvCpdSpawn:		tmp = "stvCpdSpawn";			break;
		case stvCpdAout:		tmp = "stvCpdAout";				break;
		case stvCpdUserNotifyInfo:
								tmp = "stvCpdUserNotifyInfo";	break;
		case stvCpdNproc:		tmp = "stvCpdNproc";			break;
		case stvCpdPeerFailed:	tmp = "stvCpdPeerFailed";		break;
		case stvCpdNewPeer:		tmp = "stvCpdNewPeer";			break;
		case stvCpdShutdown:	tmp = "stvCpdShutdown";			break;
		case stvCpdForwardPeerFailed:
								tmp = "stvCpdForwardPeerFailed";
																break;
		case stvCpdReconfigRqst:
								tmp = "stvCpdReconfigRqst";		break;
		case stvCpdReconfigReply:
								tmp = "stvCpdReconfigReply";	break;
		case stvCpdListFiles:
								tmp = "stvCpdListFiles";		break;
		case stvCpdFileList:
								tmp = "stvCpdFileList";			break;
		case stvCpdBootstrap:
								tmp = "stvCpdBootstrap";		break;
		case stvCpdVerifyCheckpointRelease:
								tmp = "stvCpdVerifyCheckpointRelease";
																break;
		case stvCheckpointRelease:
								tmp = "stvCheckpointRelease";	break;
		case stvAppFinished:
								tmp = "stvAppFinished";			break;
		case stvCpdVerifyAppFinished:
								tmp = "stvCpdVerifyAppFinished";
																break;
		case stvAppRelease:		tmp = "stvAppRelease";			break;

		default:
		{
			if ( id < 0 )
				tmp = "<Notify Message>";

			else
				tmp = "<Unknown Message Op>";

			break;
		}
	}

	return( tmp );
}


char *
stv_data_type_str( id )
int id;
{
	char *tmp;

	switch ( id )
	{
		case stvStr:	tmp = "String";				break;
		case stvByte:	tmp = "Byte";				break;
		case stvShort:	tmp = "Short";				break;
		case stvInt:	tmp = "Integer";			break;
		case stvFloat:	tmp = "Float";				break;
		case stvCplx:	tmp = "Complex";			break;
		case stvDouble:	tmp = "Double";				break;
		case stvDcplx:	tmp = "Double Complex";		break;
		case stvLong:	tmp = "Long";				break;
		case stvUshort:	tmp = "Unsigned Short";		break;
		case stvUint:	tmp = "Unsigned Integer";	break;
		case stvUlong:	tmp = "Unsigned Long";		break;

		default:	tmp = "<Unknown Data Type>";
	}

	return( tmp );
}


char *
stv_storage_order_str( id )
int id;
{
	char *tmp;

	switch ( id )
	{
		case stvColumnMajor:	tmp = "Column Major";	break;
		case stvRowMajor:		tmp = "Row Major";		break;

		default:	tmp = "<Unknown Data Storage Order>";
	}

	return( tmp );
}


char *
stv_decomp_type_str( id )
int id;
{
	char *tmp;

	switch ( id )
	{
		case stvBlock:		tmp = "stvBlock";		break;
		case stvCyclic:		tmp = "stvCyclic";		break;
		case stvExplicit:	tmp = "stvExplicit";	break;
		case stvCollapse:	tmp = "stvCollapse";	break;

		default:	tmp = "<Unknown Decomp Type>";
	}

	return( tmp );
}


char *
stv_option_str( id )
int id;
{
	char *tmp;

	switch ( id )
	{
		case stvOptFullUnpack:		tmp = "stvOptFullUnpack";	break;
		case stvOptVerbose:			tmp = "stvOptVerbose";		break;
		case stvOptDumpAll:			tmp = "stvOptDumpAll";		break;
		case stvOptDump:			tmp = "stvOptDump";			break;
		case stvOptFieldStorage:	tmp = "stvOptFieldStorage";	break;
		case stvOptAppLock:			tmp = "stvOptAppLock";		break;
		case stvOptCpRecovery:		tmp = "stvOptCpRecovery";	break;
		case stvOptFieldTimeScale:	tmp = "stvOptFieldTimeScale";
																break;
		case stvOptSetFieldTime:	tmp = "stvOptSetFieldTime";	break;
		case stvOptIncrFieldTime:	tmp = "stvOptIncrFieldTime";
																break;
		case stvOptFieldReady:		tmp = "stvOptFieldReady";	break;

		case stvOptAll:				tmp = "stvOptAll";			break;
		case stvOptNull:			tmp = "stvOptNull";			break;

		default:					tmp = "<Unknown Option>";	break;
	}

	return( tmp );
}


int
stv_dump_time_str( fp, tv, nlflag )
FILE *fp;
int tv;
int nlflag;
{
	int value;

	if ( fp == NULL )
	{
		STV_ERROR( "stv_dump_time_str", stvStatusBadParam,
				"Null File Pointer" );
			
		return( stvStatusBadParam );
	}

	if ( stv_is_time_value( tv ) )
	{
		value = stv_value_of_time( tv );

		fprintf( fp, "%d %s%s%s", value,
				stv_time_unit_str( stv_unit_of_time( tv ) ),
				( value != 1 ) ? "s" : "", ( nlflag ) ? "\n" : "" );
	}

	else
	{
		fprintf( fp, "%d %s%s%s",
			tv, stv_time_unit_str( stvTimeNone ),
			( tv != 1 ) ? "s" : "", ( nlflag ) ? "\n" : "" );
	}

	return( stvStatusOk );
}


int
stv_dump_base_time_str( fp, base, units, nlflag )
FILE *fp;
int base;
int units;
int nlflag;
{
	TIME_CONVERSION *TC;

	int factor;

	int first;
	int val;

	if ( fp == NULL )
	{
		STV_ERROR( "stv_dump_base_time_str", stvStatusBadParam,
				"Null File Pointer" );
			
		return( stvStatusBadParam );
	}

	if ( stv_is_time_value( units ) )
	{
		fprintf( fp, "%d %s%s",
				base, stv_time_unit_str( units ),
				( base != 1 ) ? "s" : "" );

		/* Check for Potential Time Break Down... */

		TC = &(STV_TIME_CONV[0]);

		while ( TC->units && TC->units != units )
			TC++;

		if ( TC->units )
		{
			/* Determine Max Factor */

			factor = 1;

			while ( factor <= base && TC->units )
			{
				factor *= TC->value;

				TC++;
			}

			/* Potential Factor Found */

			if ( factor > 1 )
			{
				/* Break Down Time */

				first = 1;

				do
				{
					/* Back Off By One... */

					TC--;

					factor /= TC->value;

					if ( factor <= base )
					{
						/* Print Time Component */

						if ( first )
						{
							fprintf( fp, " ( " );

							first = 0;
						}

						else
							fprintf( fp, ", " );

						val = base / factor;

						stv_dump_time_str( fp,
								stv_time_of( val, TC->units ),
								stvFalse );

						base -= val * factor;
					}
				}
				while ( factor > 1 );

				if ( !first )
					fprintf( fp, " )" );
			}
		}

		if ( nlflag )
			fprintf( fp, "\n" );
	}

	else
	{
		fprintf( fp, "%d %s%s%s",
			base, stv_time_unit_str( stvTimeNone ),
			( base != 1 ) ? "s" : "", ( nlflag ) ? "\n" : "" );
	}

	return( stvStatusOk );
}


#ifdef WIN32
int
strncasecmp( str1, str2, n )
char *str1, *str2;
int n;
{
	char *s1, *s2;

	int ret;
	int i;

	if ( str1 == NULL || str2 == NULL )
		return( -1 );

	s1 = stv_copy_str( str1 );
	s2 = stv_copy_str( str2 );

	for ( i=0 ; i < strlen( str1 ) ; i++ )
		s1[i] = toupper( str1[i] );

	for ( i=0 ; i < strlen( str2 ) ; i++ )
		s2[i] = toupper( str2[i] );

	ret = strncmp( s1, s2, n );

	return( ret );
}
#endif

int
stv_time_unit_of( str )
char *str;
{
	int tu;

	/* Check for time unit strings */

	if ( !strncasecmp( str, "pico", 4 )
			|| !strncasecmp( str, "psec", 4 ) )
		tu = stvTimePicoSec;

	else if ( !strncasecmp( str, "nano", 4 )
			|| !strncasecmp( str, "nsec", 4 ) )
		tu = stvTimeNanoSec;

	else if ( !strncasecmp( str, "micro", 5 )
			|| !strncasecmp( str, "usec", 4 ) )
		tu = stvTimeMicroSec;

	else if ( !strncasecmp( str, "milli", 5 )
			|| !strncasecmp( str, "msec", 4 ) )
		tu = stvTimeMilliSec;

	else if ( !strncasecmp( str, "sec", 3 ) )
		tu = stvTimeSec;

	else if ( !strncasecmp( str, "min", 3 ) )
		tu = stvTimeMin;

	else if ( !strncasecmp( str, "hour", 4 )
			|| !strncasecmp( str, "hr", 2 ) )
		tu = stvTimeHour;

	else if ( !strncasecmp( str, "day", 3 ) )
		tu = stvTimeDay;

	else if ( !strncasecmp( str, "week", 4 )
			|| !strncasecmp( str, "wk", 2 ) )
		tu = stvTimeWeek;

	else if ( !strncasecmp( str, "year", 4 )
			|| !strncasecmp( str, "yr", 2 ) )
		tu = stvTimeYear;

	else if ( !strncasecmp( str, "dec", 3 ) )
		tu = stvTimeDecade;

	else if ( !strncasecmp( str, "cent", 4 ) )
		tu = stvTimeCentury;

	/* default */
	else
	{
		printf( "%s: Unknown Time Unit String \"%s\"...\n",
			"stv_time_unit_of()", str );

		tu = stvNotTime;
	}

	return( tu );
}


char *
stv_time_unit_str( unit )
int unit;
{
	char *tmp;

	switch( unit )
	{
		case stvTimePicoSec:	tmp = "PicoSec";	break;
		case stvTimeNanoSec:	tmp = "NanoSec";	break;
		case stvTimeMicroSec:	tmp = "MicroSec";	break;
		case stvTimeMilliSec:	tmp = "MilliSec";	break;
		case stvTimeSec:		tmp = "Sec";		break;
		case stvTimeMin:		tmp = "Min";		break;
		case stvTimeHour:		tmp = "Hour";		break;
		case stvTimeDay:		tmp = "Day";		break;
		case stvTimeWeek:		tmp = "Week";		break;
		case stvTimeYear:		tmp = "Year";		break;
		case stvTimeDecade:		tmp = "Decade";		break;
		case stvTimeCentury:	tmp = "Century";	break;

		case stvTimeNone:		tmp = "Iteration";	break;

		default:				tmp = "<Unknown Time Units>";
	}

	return( tmp );
}


void
stv_dump_time_constants()
{
	printf( "stvTimeNegative = %u / %d / %x\n",
			stvTimeNegative, stvTimeNegative, stvTimeNegative );
	printf( "stvTimeBase = %u / %d / %x\n",
			stvTimeBase, stvTimeBase, stvTimeBase );
	printf( "stvTimeShift = %u / %d / %x\n",
			stvTimeShift, stvTimeShift, stvTimeShift );
	printf( "stvTimeMaxValue = %u / %d / %x\n",
			stvTimeMaxValue, stvTimeMaxValue, stvTimeMaxValue );
	printf( "(~stvTimeMaxValue) = %u / %d / %x\n",
			(~stvTimeMaxValue), (~stvTimeMaxValue),
			(~stvTimeMaxValue) );
	printf( "(stvTimeNegative - 1) = %u / %d / %x\n",
			(stvTimeNegative - 1), (stvTimeNegative - 1),
			(stvTimeNegative - 1) );
	printf(
		"(~stvTimeMaxValue) & (stvTimeNegative - 1) = %u / %d / %x\n",
			(~stvTimeMaxValue) & (stvTimeNegative - 1),
			(~stvTimeMaxValue) & (stvTimeNegative - 1),
			(~stvTimeMaxValue) & (stvTimeNegative - 1) );
	printf( "\n" );

	printf( "stvTimePicoSec = %u / %d / %x\n",
			stvTimePicoSec, stvTimePicoSec, stvTimePicoSec );
	printf( "stvTimeNanoSec = %u / %d / %x\n",
			stvTimeNanoSec, stvTimeNanoSec, stvTimeNanoSec );
	printf( "stvTimeMicroSec = %u / %d / %x\n",
			stvTimeMicroSec, stvTimeMicroSec, stvTimeMicroSec );
	printf( "stvTimeMilliSec = %u / %d / %x\n",
			stvTimeMilliSec, stvTimeMilliSec, stvTimeMilliSec );
	printf( "stvTimeSec = %u / %d / %x\n",
			stvTimeSec, stvTimeSec, stvTimeSec );
	printf( "stvTimeMin = %u / %d / %x\n",
			stvTimeMin, stvTimeMin, stvTimeMin );
	printf( "stvTimeHour = %u / %d / %x\n",
			stvTimeHour, stvTimeHour, stvTimeHour );
	printf( "stvTimeDay = %u / %d / %x\n",
			stvTimeDay, stvTimeDay, stvTimeDay );
	printf( "stvTimeWeek = %u / %d / %x\n",
			stvTimeWeek, stvTimeWeek, stvTimeWeek );
	printf( "stvTimeYear = %u / %d / %x\n",
			stvTimeYear, stvTimeYear, stvTimeYear );
	printf( "stvTimeDecade = %u / %d / %x\n",
			stvTimeDecade, stvTimeDecade, stvTimeDecade );
	printf( "stvTimeCentury = %u / %d / %x\n",
			stvTimeCentury, stvTimeCentury, stvTimeCentury );
	printf( "\n" );
}


void
stvSleep( t )
int t;
{
#ifdef WIN32

	Sleep( t );

#else

	sleep( t );

#endif
}


char *
stv_copy_str( str )
char *str;
{
	char *tmp;

	tmp = (char *) MALLOC( (STV_SIZE_T) ( strlen( str ) + 1 )
		* sizeof(char) );
	STV_MEMCHECK( tmp, "Copy String", return( (char *) NULL ) );

	strcpy( tmp, str );

	return( tmp );
}


int
stv_memcheck( ptr, name )
void *ptr;
char *name;
{
	if ( ptr == NULL )
	{
		sprintf( STVERRTXT, "Allocating Memory for \"%s\"", name );

		STV_ERROR( "stv_memcheck", stvStatusNoMem, STVERRTXT );

		return( stvStatusNoMem );
	}

	return( stvStatusOk );
}


int
stv_pack_by_type( bufid, val, cnt, stride, type )
int bufid;
STV_VALUE val;
int cnt, stride, type;
{
	int cc;

	switch ( type )
	{
		case stvByte:
			cc = stv_pack_byte( bufid, val, cnt, stride );
			break;
		case stvStr:
			cc = stv_pack_str( bufid, val );
			break;
		case stvInt:
			cc = stv_pack_int( bufid, val, cnt, stride );
			break;
		case stvLong:
			cc = stv_pack_long( bufid, val, cnt, stride );
			break;
		case stvFloat:
			cc = stv_pack_float( bufid, val, cnt, stride );
			break;
		case stvDouble:
			cc = stv_pack_double( bufid, val, cnt, stride );
			break;
		case stvCplx:
			cc = stv_pack_complex( bufid, val, cnt, stride );
			break;
		case stvDcplx:
			cc = stv_pack_dcomplex( bufid, val, cnt, stride );
			break;
		case stvUshort:
			cc = stv_pack_ushort( bufid, val, cnt, stride );
			break;
		case stvUint:
			cc = stv_pack_uint( bufid, val, cnt, stride );
			break;
		case stvUlong:
			cc = stv_pack_ulong( bufid, val, cnt, stride );
			break;
		default:
			sprintf( STVERRTXT, "Unknown Data Type = %d", type );
			STV_ERROR( "stv_pack_by_type", stvStatusBadParam,
				STVERRTXT );
			cc = stvStatusBadParam;
			break;
	}

	cc = cc ? stvStatusOk : stvStatusBadParam;

	return( cc );
}


int
stv_unpack_by_type( bufid, val, cnt, stride, type )
int bufid;
STV_VALUE val;
int cnt, stride, type;
{
	char tmpstr[1024];	/* much bad wind */

	int cc;

	switch ( type )
	{
		case stvByte:
			cc = stv_unpack_byte( bufid, val, cnt, stride );
			break;
		case stvStr:
			cc = stv_unpack_str( bufid, tmpstr );
			*((char **) val) = stv_copy_str( tmpstr );
			/* never use this... :-) */
			break;
		case stvInt:
			cc = stv_unpack_int( bufid, val, cnt, stride );
			break;
		case stvLong:
			cc = stv_unpack_long( bufid, val, cnt, stride );
			break;
		case stvFloat:
			cc = stv_unpack_float( bufid, val, cnt, stride );
			break;
		case stvDouble:
			cc = stv_unpack_double( bufid, val, cnt, stride );
			break;
		case stvCplx:
			cc = stv_unpack_complex( bufid, val, cnt, stride );
			break;
		case stvDcplx:
			cc = stv_unpack_dcomplex( bufid, val, cnt, stride );
			break;
		case stvUshort:
			cc = stv_unpack_ushort( bufid, val, cnt, stride );
			break;
		case stvUint:
			cc = stv_unpack_uint( bufid, val, cnt, stride );
			break;
		case stvUlong: 
			cc = stv_unpack_ulong( bufid, val, cnt, stride ); 
			break;
		default:
			sprintf( STVERRTXT, "Unknown Data Type = %d", type );
			STV_ERROR( "stv_unpack_by_type", stvStatusBadParam,
				STVERRTXT );
			cc = stvStatusBadParam;
			break;
	}

	cc = cc ? stvStatusOk : stvStatusBadParam;

	return( cc );
}


int
stv_copy_converted_data_elem( src, src_type, src_index,
		dst, dst_type, dst_index )
STV_VALUE src;
int src_type;
int src_index;
STV_VALUE dst;
int dst_type;
int dst_index;
{
	return( stv_copy_converted_data(
			(STV_VALUE) ( (char *) src
				+ ( (STV_SIZE_T) src_index * STV_SIZEOF( src_type ) ) ),
			src_type,
			(STV_VALUE) ( (char *) dst
				+ ( (STV_SIZE_T) dst_index * STV_SIZEOF( dst_type ) ) ),
			dst_type ) );
}


int
stv_copy_converted_data( src, src_type, dst, dst_type )
STV_VALUE src;
int src_type;
STV_VALUE dst;
int dst_type;
{
	double tmp;
	double tmp2;

	/* Convert Source Value to Double(s) */

	if ( dst_type != stvStr )
	{
		tmp2 = 0.0;

		switch ( src_type )
		{
			case stvStr:
				tmp = 0.0;
				break;
			case stvByte:
				tmp = (double) STV_VALUE_OF( src, char );
				break;
			case stvInt:
				tmp = (double) STV_VALUE_OF( src, int );
				break;
			case stvLong:
				tmp = (double) STV_VALUE_OF( src, long );
				break;
			case stvFloat:
				tmp = (double) STV_VALUE_OF( src, float );
				break;
			case stvDouble:
				tmp = STV_VALUE_OF( src, double );
				break;
			case stvCplx:
				tmp = (double) STV_VALUE_OF( src, float );
				tmp2 = (double) STV_VALUE_OF( src + 1, float );
				break;
			case stvDcplx:
				tmp = STV_VALUE_OF( src, double );
				tmp2 = STV_VALUE_OF( src + 1, double );
				break;
			case stvUshort:
				tmp = (double) STV_VALUE_OF( src, unsigned short );
				break;
			case stvUint:
				tmp = (double) STV_VALUE_OF( src, unsigned int );
				break;
			case stvUlong:
				tmp = (double) STV_VALUE_OF( src, unsigned long );
				break;
			default:
				sprintf( STVERRTXT, "Unknown Data Type = %d",
					src_type );
				STV_ERROR( "stv_copy_converted_data", stvStatusBadParam,
					STVERRTXT );
				return( stvStatusBadParam );
				/* break; */
		}

		switch ( dst_type )
		{
			/* no case stvStr: handled below */
			case stvByte:
				STV_VALUE_OF( dst, char ) = (char) tmp;
				break;
			case stvInt:
				STV_VALUE_OF( dst, int ) = (int) tmp;
				break;
			case stvLong:
				STV_VALUE_OF( dst, long ) = (long) tmp;
				break;
			case stvFloat:
				STV_VALUE_OF( dst, float ) = (float) tmp;
				break;
			case stvDouble:
				STV_VALUE_OF( dst, double ) = tmp;
				break;
			case stvCplx:
				STV_VALUE_OF( dst, float ) = (float) tmp;
				STV_VALUE_OF( dst + 1, float ) = (float) tmp2;
				break;
			case stvDcplx:
				STV_VALUE_OF( dst, double ) = tmp;
				STV_VALUE_OF( dst + 1, double ) = tmp2;
				break;
			case stvUshort:
				STV_VALUE_OF( dst, unsigned short ) =
					(unsigned short) tmp;
				break;
			case stvUint:
				STV_VALUE_OF( dst, unsigned int ) =
					(unsigned int) tmp;
				break;
			case stvUlong:
				STV_VALUE_OF( dst, unsigned long ) =
					(unsigned long) tmp;
				break;
			default:
				sprintf( STVERRTXT, "Unknown Data Type = %d",
					dst_type );
				STV_ERROR( "stv_copy_converted_data", stvStatusBadParam,
					STVERRTXT );
				return( stvStatusBadParam );
				/* break; */
		}
	}

	else
	{
		STV_VALUE_OF( dst, char * ) =
			stv_str_of_typed_data( src, src_type );

		if ( STV_VALUE_OF( dst, char * ) == NULL )
		{
			STV_ERROR( "stv_copy_converted_data", stvStatusBadParam,
				"stv_str_of_typed_data() failed." );

			return( stvStatusBadParam );
		}
	}

	return( stvStatusOk );
}


char *
stv_str_of_typed_data( ptr, type )
STV_VALUE ptr;
int type;
{
	char tmpstr[1024];

	switch ( type )
	{
		case stvStr:
			sprintf( tmpstr, "%s", STV_VALUE_OF( ptr, char * ) );
			break;
		case stvByte:
			sprintf( tmpstr, "%d", STV_VALUE_OF( ptr, char ) );
			break;
		case stvInt:
			sprintf( tmpstr, "%d", STV_VALUE_OF( ptr, int ) );
			break;
		case stvLong:
			sprintf( tmpstr, "%ld", STV_VALUE_OF( ptr, long ) );
			break;
		case stvFloat:
			sprintf( tmpstr, "%g", STV_VALUE_OF( ptr, float ) );
			break;
		case stvDouble:
			sprintf( tmpstr, "%lg", STV_VALUE_OF( ptr, double ) );
			break;
		case stvCplx:
			sprintf( tmpstr, "(%g,%g)",
				STV_VALUE_OF( ptr, float ),
				STV_VALUE_OF( ptr + 1, float ) );
			break;
		case stvDcplx:
			sprintf( tmpstr, "(%lg,%lg)",
				STV_VALUE_OF( ptr, double ),
				STV_VALUE_OF( ptr + 1, double ) );
			break;
		case stvUshort:
			sprintf( tmpstr, "%u",
				STV_VALUE_OF( ptr, unsigned short ) );
			break;
		case stvUint:
			sprintf( tmpstr, "%u",
				STV_VALUE_OF( ptr, unsigned int ) );
			break;
		case stvUlong:
			sprintf( tmpstr, "%lu",
				STV_VALUE_OF( ptr, unsigned long ) );
			break;
		default:
			sprintf( STVERRTXT, "Unknown Data Type = %d", type );
			STV_ERROR( "stv_str_of_typed_data", stvStatusBadParam,
				STVERRTXT );
			return( (char *) NULL );
			/* break; */
	}

	return( stv_copy_str( tmpstr ) );
}


STV_VALUE
stv_get_typed_ptr_offset( ptr, offset, type )
STV_VALUE ptr;
int offset;
int type;
{
	STV_VALUE new;

	new = (STV_VALUE) NULL;

	switch ( type )
	{
		case stvStr:
			new = STV_PTR_OF( ptr, char * ) + offset;
			break;
		case stvByte:
			new = STV_PTR_OF( ptr, char ) + offset;
			break;
		case stvInt:
			new = STV_PTR_OF( ptr, int ) + offset;
			break;
		case stvLong:
			new = STV_PTR_OF( ptr, long ) + offset;
			break;
		case stvFloat:
			new = STV_PTR_OF( ptr, float ) + offset;
			break;
		case stvDouble:
			new = STV_PTR_OF( ptr, double ) + offset;
			break;
		case stvCplx:
			new = STV_PTR_OF( ptr, float ) + ( 2 * offset );
			break;
		case stvDcplx:
			new = STV_PTR_OF( ptr, double ) + ( 2 * offset );
			break;
		case stvUshort:
			new = STV_PTR_OF( ptr, unsigned short ) + offset;
			break;
		case stvUint:
			new = STV_PTR_OF( ptr, unsigned int ) + offset;
			break;
		case stvUlong:
			new = STV_PTR_OF( ptr, unsigned long ) + offset;
			break;
		default:
			sprintf( STVERRTXT, "Unknown Data Type = %d", type );
			STV_ERROR( "stv_get_typed_ptr_offset", stvStatusBadParam,
				STVERRTXT );
			break;
	}

	return( new );
}

